#! /usr/bin/env python
"""
A Command Line Wrapper to allow easy use of pyicloud for
command line scripts, and related.
"""

import argparse
import logging
import os
import pickle
import sys
from pprint import pformat
from typing import Any, Optional

from click import confirm

from pyicloud import PyiCloudService, utils
from pyicloud.exceptions import PyiCloudFailedLoginException, PyiCloudServiceUnavailable
from pyicloud.services.findmyiphone import AppleDevice
from pyicloud.ssl_context import configurable_ssl_verification

DEVICE_ERROR = "Please use the --device switch to indicate which device to use."


def create_pickled_data(idevice: AppleDevice, filename: str) -> None:
    """
    This helper will output the idevice to a pickled file named
    after the passed filename.

    This allows the data to be used without resorting to screen / pipe
    scrapping.
    """
    with open(filename, "wb") as pickle_file:
        pickle.dump(
            idevice.data,
            pickle_file,
            protocol=pickle.HIGHEST_PROTOCOL,
        )


def _create_parser() -> argparse.ArgumentParser:
    """Create the parser."""
    parser = argparse.ArgumentParser(description="Find My iPhone CommandLine Tool")

    parser.add_argument(
        "--username",
        action="store",
        dest="username",
        default="",
        help="Apple ID to Use",
    )

    parser.add_argument(
        "--password",
        action="store",
        dest="password",
        default="",
        help=(
            "Apple ID Password to Use; if unspecified, password will be "
            "fetched from the system keyring."
        ),
    )

    parser.add_argument(
        "--china-mainland",
        action="store_true",
        dest="china_mainland",
        default=False,
        help="If the country/region setting of the Apple ID is China mainland",
    )

    parser.add_argument(
        "-n",
        "--non-interactive",
        action="store_false",
        dest="interactive",
        default=True,
        help="Disable interactive prompts.",
    )

    parser.add_argument(
        "--delete-from-keyring",
        action="store_true",
        dest="delete_from_keyring",
        default=False,
        help="Delete stored password in system keyring for this username.",
    )

    # Group for listing options
    list_group = parser.add_argument_group(
        title="Listing Options",
        description="Options for listing devices",
    )

    # Mutually exclusive group for listing options
    list_type_group = list_group.add_mutually_exclusive_group()
    list_type_group.add_argument(
        "--list",
        action="store_true",
        dest="list",
        default=False,
        help="Short Listings for Device(s) associated with account",
    )

    list_type_group.add_argument(
        "--llist",
        action="store_true",
        dest="longlist",
        default=False,
        help="Detailed Listings for Device(s) associated with account",
    )

    list_group.add_argument(
        "--locate",
        action="store_true",
        dest="locate",
        default=False,
        help="Retrieve Location for the iDevice (non-exclusive).",
    )

    # Restrict actions to a specific devices UID / DID
    parser.add_argument(
        "--device",
        action="store",
        dest="device_id",
        default=False,
        help="Only effect this device",
    )

    # Trigger Sound Alert
    parser.add_argument(
        "--sound",
        action="store_true",
        dest="sound",
        default=False,
        help="Play a sound on the device",
    )

    # Trigger Message w/Sound Alert
    parser.add_argument(
        "--message",
        action="store",
        dest="message",
        default=False,
        help="Optional Text Message to display with a sound",
    )

    # Trigger Message (without Sound) Alert
    parser.add_argument(
        "--silentmessage",
        action="store",
        dest="silentmessage",
        default=False,
        help="Optional Text Message to display with no sounds",
    )

    # Lost Mode
    parser.add_argument(
        "--lostmode",
        action="store_true",
        dest="lostmode",
        default=False,
        help="Enable Lost mode for the device",
    )

    parser.add_argument(
        "--lostphone",
        action="store",
        dest="lost_phone",
        default=False,
        help="Phone Number allowed to call when lost mode is enabled",
    )

    parser.add_argument(
        "--lostpassword",
        action="store",
        dest="lost_password",
        default=False,
        help="Forcibly active this passcode on the idevice",
    )

    parser.add_argument(
        "--lostmessage",
        action="store",
        dest="lost_message",
        default="",
        help="Forcibly display this message when activating lost mode.",
    )

    # Output device data to an pickle file
    parser.add_argument(
        "--outputfile",
        action="store_true",
        dest="output_to_file",
        default="",
        help="Save device data to a file in the current directory.",
    )

    parser.add_argument(
        "--log-level",
        action="store",
        dest="loglevel",
        choices=["error", "warning", "info", "none"],
        default="info",
        help="Set the logging level",
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug logging",
    )

    parser.add_argument(
        "--accept-terms",
        action="store_true",
        default=False,
        help="Automatically accept terms and conditions",
    )

    parser.add_argument(
        "--with-family",
        action="store_true",
        default=False,
        help="Include family devices",
    )

    parser.add_argument(
        "--session-dir",
        type=str,
        help="Directory to store session files",
    )

    parser.add_argument(
        "--http-proxy",
        type=str,
        help="Use HTTP proxy for requests",
    )

    parser.add_argument(
        "--https-proxy",
        type=str,
        help="Use HTTPS proxy for requests",
    )

    parser.add_argument(
        "--no-verify-ssl",
        action="store_true",
        default=False,
        help="Disable SSL certificate verification (WARNING: This makes the connection insecure)",
    )

    return parser


def _get_password(
    username: str,
    parser: argparse.ArgumentParser,
    command_line: argparse.Namespace,
) -> Optional[str]:
    """Which password we use is determined by your username, so we
    do need to check for this first and separately."""
    if not username:
        parser.error("No username supplied")

    password: Optional[str] = command_line.password
    if not password:
        password = utils.get_password(username, interactive=command_line.interactive)

    return password


def main() -> None:
    """Main commandline entrypoint."""
    parser: argparse.ArgumentParser = _create_parser()
    command_line: argparse.Namespace = parser.parse_args()
    level = logging.INFO

    if command_line.loglevel == "error":
        level = logging.ERROR
    elif command_line.loglevel == "warning":
        level = logging.WARNING
    elif command_line.loglevel == "info":
        level = logging.INFO
    elif command_line.loglevel == "none":
        level = None

    if command_line.debug:
        level = logging.DEBUG

    if level:
        logging.basicConfig(level=level)

    username: str = command_line.username.strip()
    china_mainland: bool = command_line.china_mainland

    if username and command_line.delete_from_keyring:
        utils.delete_password_in_keyring(username)

    with configurable_ssl_verification(
        verify_ssl=not command_line.no_verify_ssl,
        http_proxy=command_line.http_proxy or "",
        https_proxy=command_line.https_proxy or "",
    ):
        password: Optional[str] = _get_password(username, parser, command_line)

        api: Optional[PyiCloudService] = _authenticate(
            username,
            password,
            china_mainland,
            parser,
            command_line,
        )

        if not api:
            return
        _print_devices(api, command_line)


def _authenticate(
    username: str,
    password: Optional[str],
    china_mainland: bool,
    parser: argparse.ArgumentParser,
    command_line: argparse.Namespace,
) -> Optional[PyiCloudService]:
    api = None
    try:
        api = PyiCloudService(
            apple_id=username,
            password=password,
            china_mainland=china_mainland,
            cookie_directory=command_line.session_dir,
            accept_terms=command_line.accept_terms,
            with_family=command_line.with_family,
        )
        if (
            not utils.password_exists_in_keyring(username)
            and command_line.interactive
            and confirm("Save password in keyring?")
            and password
        ):
            utils.store_password_in_keyring(username, password)

        if api.requires_2fa:
            _handle_2fa(api)

        elif api.requires_2sa:
            _handle_2sa(api)
        return api
    except PyiCloudFailedLoginException as err:
        # If they have a stored password; we just used it and
        # it did not work; let's delete it if there is one.
        if not password:
            parser.error("No password supplied")

        if utils.password_exists_in_keyring(username):
            utils.delete_password_in_keyring(username)

        message: str = f"Bad username or password for {username}"

        print(err, file=sys.stderr)

        raise RuntimeError(message) from err


def _print_devices(api: PyiCloudService, command_line: argparse.Namespace) -> None:
    try:
        print(f"Number of devices: {len(api.devices)}", flush=True)
        for dev in api.devices:
            if not command_line.device_id or (
                command_line.device_id.strip().lower() == dev.id.strip().lower()
            ):
                # List device(s)
                _list_devices_option(command_line, dev)

                # Play a Sound on a device
                _play_device_sound_option(command_line, dev)

                # Display a Message on the device
                _display_device_message_option(command_line, dev)

                # Display a Silent Message on the device
                _display_device_silent_message_option(command_line, dev)

                # Enable Lost mode
                _enable_lost_mode_option(command_line, dev)
    except PyiCloudServiceUnavailable:
        print("iCloud - Find My service is unavailable.")


def _enable_lost_mode_option(
    command_line: argparse.Namespace, dev: AppleDevice
) -> None:
    if command_line.lostmode:
        if command_line.device_id:
            dev.lost_device(
                number=command_line.lost_phone.strip(),
                text=command_line.lost_message.strip(),
                newpasscode=command_line.lost_password.strip(),
            )
        else:
            raise RuntimeError(
                f"Lost Mode can only be activated on a singular device. {DEVICE_ERROR}"
            )


def _display_device_silent_message_option(
    command_line: argparse.Namespace, dev: AppleDevice
) -> None:
    if command_line.silentmessage:
        if command_line.device_id:
            dev.display_message(
                subject="A Silent Message",
                message=command_line.silentmessage,
                sounds=False,
            )
        else:
            raise RuntimeError(
                f"Silent Messages can only be played on a singular device. {DEVICE_ERROR}"
            )


def _display_device_message_option(
    command_line: argparse.Namespace, dev: AppleDevice
) -> None:
    if command_line.message:
        if command_line.device_id:
            dev.display_message(
                subject="A Message", message=command_line.message, sounds=True
            )
        else:
            raise RuntimeError(
                f"Messages can only be played on a singular device. {DEVICE_ERROR}"
            )


def _play_device_sound_option(
    command_line: argparse.Namespace, dev: AppleDevice
) -> None:
    if command_line.sound:
        if command_line.device_id:
            dev.play_sound()
        else:
            raise RuntimeError(
                f"\n\n\t\tSounds can only be played on a singular device. {DEVICE_ERROR}\n\n"
            )


def _list_devices_option(command_line: argparse.Namespace, dev: AppleDevice) -> None:
    location = dev.location if command_line.locate else None

    if command_line.output_to_file:
        create_pickled_data(
            dev,
            filename=(dev.name.strip().lower() + ".fmip_snapshot"),
        )

    if command_line.longlist:
        print("-" * 30)
        print(dev.name)
        for key in dev.data:
            print(
                f"{key:>30} - {pformat(dev.data[key]).replace(os.linesep, os.linesep + ' ' * 33)}"
            )
    elif command_line.list:
        print("-" * 30)
        print(f"Name           - {dev.name}")
        print(f"Display Name   - {dev.deviceDisplayName}")
        print(f"Location       - {location or dev.location}")
        print(f"Battery Level  - {dev.batteryLevel}")
        print(f"Battery Status - {dev.batteryStatus}")
        print(f"Device Class   - {dev.deviceClass}")
        print(f"Device Model   - {dev.deviceModel}")


def _handle_2fa(api: PyiCloudService) -> None:
    print("\nTwo-step authentication required.", "\nPlease enter validation code")

    code: str = input("(string) --> ")
    if not api.validate_2fa_code(code):
        print("Failed to verify verification code")
        sys.exit(1)

    print("")


def _handle_2sa(api: PyiCloudService) -> None:
    print("\nTwo-step authentication required.", "\nYour trusted devices are:")

    devices: list[dict[str, Any]] = _show_devices(api)

    print("\nWhich device would you like to use?")
    device_idx = int(input("(number) --> "))
    device: dict[str, Any] = devices[device_idx]
    if not api.send_verification_code(device):
        print("Failed to send verification code")
        sys.exit(1)

    print("\nPlease enter validation code")
    code: str = input("(string) --> ")
    if not api.validate_verification_code(device, code):
        print("Failed to verify verification code")
        sys.exit(1)

    print("")


def _show_devices(api: PyiCloudService) -> list[dict[str, Any]]:
    """Show devices."""
    devices: list[dict[str, Any]] = api.trusted_devices
    for i, device in enumerate(devices):
        phone_number: str = f"{device.get('deviceType')} to {device.get('phoneNumber')}"
        print(f"    {i}: {device.get('deviceName', phone_number)}")

    return devices


if __name__ == "__main__":
    main()



"""Login test constants."""

from typing import Any

from .const_account_family import (
    APPLE_ID_EMAIL,
    FIRST_NAME,
    FULL_NAME,
    ICLOUD_ID_EMAIL,
    LAST_NAME,
    PERSON_ID,
    PRIMARY_EMAIL,
)

NOTIFICATION_ID: str = "12345678-1234-1234-1234-123456789012" + PERSON_ID
A_DS_ID: str = "123456-12-12345678-1234-1234-1234-123456789012" + PERSON_ID
WIDGET_KEY: str = "widget_key" + PERSON_ID

# Data
AUTH_OK: dict[str, Any] = {
    "authType": "hsa2",
    "salt": "U29tZVNhbHQ=",
    "b": "U29tZUJ5dGVz",
    "c": "TestC",
    "iteration": 1000,
    "protocol": "s2k",
    "dsInfo": {"hsaVersion": 1},
    "hsaChallengeRequired": False,
    "webservices": "TestWebservices",
}
ICLOUD_AUTH_URL = "https://idmsa.apple.com/appleauth/auth"
ICLOUD_UPLOAD_PHOTOS_WS_URL = "https://p31-uploadphotosws.icloud.com:443"
ICLOUD_WIDGET_ACCOUNT_URL = "https://appleid.apple.com/widget/account/?widgetKey="
LOGIN_WORKING: dict[str, Any] = {
    "dsInfo": {
        "lastName": LAST_NAME,
        "iCDPEnabled": False,
        "tantorMigrated": True,
        "dsid": PERSON_ID,
        "hsaEnabled": True,
        "ironcadeMigrated": True,
        "locale": "fr-fr_FR",
        "brZoneConsolidated": False,
        "isManagedAppleID": False,
        "gilligan-invited": "true",
        "appleIdAliases": [APPLE_ID_EMAIL, ICLOUD_ID_EMAIL],
        "hsaVersion": 2,
        "isPaidDeveloper": False,
        "countryCode": "FRA",
        "notificationId": NOTIFICATION_ID,
        "primaryEmailVerified": True,
        "aDsID": A_DS_ID,
        "locked": False,
        "hasICloudQualifyingDevice": True,
        "primaryEmail": PRIMARY_EMAIL,
        "appleIdEntries": [
            {"isPrimary": True, "type": "EMAIL", "value": PRIMARY_EMAIL},
            {"type": "EMAIL", "value": APPLE_ID_EMAIL},
            {"type": "EMAIL", "value": ICLOUD_ID_EMAIL},
        ],
        "gilligan-enabled": "true",
        "fullName": FULL_NAME,
        "languageCode": "fr-fr",
        "appleId": PRIMARY_EMAIL,
        "firstName": FIRST_NAME,
        "iCloudAppleIdAlias": ICLOUD_ID_EMAIL,
        "notesMigrated": True,
        "hasPaymentInfo": False,
        "pcsDeleted": False,
        "appleIdAlias": APPLE_ID_EMAIL,
        "brMigrated": True,
        "statusCode": 2,
        "familyEligible": True,
    },
    "hasMinimumDeviceForPhotosWeb": True,
    "iCDPEnabled": False,
    "webservices": {
        "reminders": {
            "url": "https://p31-remindersws.icloud.com:443",
            "status": "active",
        },
        "notes": {"url": "https://p38-notesws.icloud.com:443", "status": "active"},
        "mail": {"url": "https://p38-mailws.icloud.com:443", "status": "active"},
        "ckdatabasews": {
            "pcsRequired": True,
            "url": "https://p31-ckdatabasews.icloud.com:443",
            "status": "active",
        },
        "photosupload": {
            "pcsRequired": True,
            "url": ICLOUD_UPLOAD_PHOTOS_WS_URL,
            "status": "active",
        },
        "photos": {
            "pcsRequired": True,
            "uploadUrl": ICLOUD_UPLOAD_PHOTOS_WS_URL,
            "url": "https://p31-photosws.icloud.com:443",
            "status": "active",
        },
        "drivews": {
            "pcsRequired": True,
            "url": "https://p31-drivews.icloud.com:443",
            "status": "active",
        },
        "uploadimagews": {
            "url": "https://p31-uploadimagews.icloud.com:443",
            "status": "active",
        },
        "schoolwork": {},
        "cksharews": {"url": "https://p31-ckshare.icloud.com:443", "status": "active"},
        "findme": {"url": "https://p31-fmipweb.icloud.com:443", "status": "active"},
        "premiummailsettings": {
            "url": "https://p42-maildomainws.icloud.com:443",
            "status": "active",
        },
        "ckdeviceservice": {"url": "https://p31-ckdevice.icloud.com:443"},
        "iworkthumbnailws": {
            "url": "https://p31-iworkthumbnailws.icloud.com:443",
            "status": "active",
        },
        "calendar": {
            "url": "https://p31-calendarws.icloud.com:443",
            "status": "active",
        },
        "docws": {
            "pcsRequired": True,
            "url": "https://p31-docws.icloud.com:443",
            "status": "active",
        },
        "settings": {
            "url": "https://p31-settingsws.icloud.com:443",
            "status": "active",
        },
        "ubiquity": {
            "url": "https://p31-ubiquityws.icloud.com:443",
            "status": "active",
        },
        "streams": {"url": "https://p31-streams.icloud.com:443", "status": "active"},
        "keyvalue": {
            "url": "https://p31-keyvalueservice.icloud.com:443",
            "status": "active",
        },
        "archivews": {
            "url": "https://p31-archivews.icloud.com:443",
            "status": "active",
        },
        "push": {"url": "https://p31-pushws.icloud.com:443", "status": "active"},
        "iwmb": {"url": "https://p31-iwmb.icloud.com:443", "status": "active"},
        "iworkexportws": {
            "url": "https://p31-iworkexportws.icloud.com:443",
            "status": "active",
        },
        "geows": {"url": "https://p31-geows.icloud.com:443", "status": "active"},
        "account": {
            "iCloudEnv": {"shortId": "p", "vipSuffix": "prod"},
            "url": "https://p31-setup.icloud.com:443",
            "status": "active",
        },
        "fmf": {"url": "https://p31-fmfweb.icloud.com:443", "status": "active"},
        "contacts": {
            "url": "https://p31-contactsws.icloud.com:443",
            "status": "active",
        },
    },
    "pcsEnabled": True,
    "configBag": {
        "urls": {
            "accountCreateUI": ICLOUD_WIDGET_ACCOUNT_URL + WIDGET_KEY + "#!create",
            "accountLoginUI": f"{ICLOUD_AUTH_URL}/signin?widgetKey={WIDGET_KEY}",
            "accountLogin": "https://setup.icloud.com/setup/ws/1/accountLogin",
            "accountRepairUI": ICLOUD_WIDGET_ACCOUNT_URL + WIDGET_KEY + "#!repair",
            "downloadICloudTerms": "https://setup.icloud.com/setup/ws/1/downloadLiteTerms",
            "repairDone": "https://setup.icloud.com/setup/ws/1/repairDone",
            "accountAuthorizeUI": f"{ICLOUD_AUTH_URL}/authorize/signin?client_id={WIDGET_KEY}",
            "vettingUrlForEmail": "https://id.apple.com/IDMSEmailVetting/vetShareEmail",
            "accountCreate": "https://setup.icloud.com/setup/ws/1/createLiteAccount",
            "getICloudTerms": "https://setup.icloud.com/setup/ws/1/getTerms",
            "vettingUrlForPhone": "https://id.apple.com/IDMSEmailVetting/vetSharePhone",
        },
        "accountCreateEnabled": "true",
    },
    "hsaTrustedBrowser": True,
    "appsOrder": [
        "mail",
        "contacts",
        "calendar",
        "photos",
        "iclouddrive",
        "notes3",
        "reminders",
        "pages",
        "numbers",
        "keynote",
        "newspublisher",
        "fmf",
        "find",
        "settings",
    ],
    "version": 2,
    "isExtendedLogin": True,
    "pcsServiceIdentitiesIncluded": True,
    "hsaChallengeRequired": False,
    "ICDRSCapableDeviceCount": 1,
    "requestInfo": {"country": "FR", "timeZone": "GMT+1", "region": "IDF"},
    "pcsDeleted": False,
    "iCloudInfo": {"SafariBookmarksHasMigratedToCloudKit": True},
    "apps": {
        "calendar": {},
        "reminders": {},
        "keynote": {"isQualifiedForBeta": True},
        "settings": {"canLaunchWithOneFactor": True},
        "mail": {},
        "numbers": {"isQualifiedForBeta": True},
        "photos": {},
        "pages": {"isQualifiedForBeta": True},
        "notes3": {},
        "find": {"canLaunchWithOneFactor": True},
        "iclouddrive": {},
        "newspublisher": {"isHidden": True},
        "fmf": {},
        "contacts": {},
    },
}

# Setup data
LOGIN_2FA = {
    "dsInfo": {
        "lastName": LAST_NAME,
        "iCDPEnabled": False,
        "tantorMigrated": True,
        "dsid": PERSON_ID,
        "hsaEnabled": True,
        "ironcadeMigrated": True,
        "locale": "fr-fr_FR",
        "brZoneConsolidated": False,
        "isManagedAppleID": False,
        "gilligan-invited": "true",
        "appleIdAliases": [APPLE_ID_EMAIL, ICLOUD_ID_EMAIL],
        "hsaVersion": 2,
        "isPaidDeveloper": False,
        "countryCode": "FRA",
        "notificationId": NOTIFICATION_ID,
        "primaryEmailVerified": True,
        "aDsID": A_DS_ID,
        "locked": False,
        "hasICloudQualifyingDevice": True,
        "primaryEmail": PRIMARY_EMAIL,
        "appleIdEntries": [
            {"isPrimary": True, "type": "EMAIL", "value": PRIMARY_EMAIL},
            {"type": "EMAIL", "value": APPLE_ID_EMAIL},
            {"type": "EMAIL", "value": ICLOUD_ID_EMAIL},
        ],
        "gilligan-enabled": "true",
        "fullName": FULL_NAME,
        "languageCode": "fr-fr",
        "appleId": PRIMARY_EMAIL,
        "firstName": FIRST_NAME,
        "iCloudAppleIdAlias": ICLOUD_ID_EMAIL,
        "notesMigrated": True,
        "hasPaymentInfo": True,
        "pcsDeleted": False,
        "appleIdAlias": APPLE_ID_EMAIL,
        "brMigrated": True,
        "statusCode": 2,
        "familyEligible": True,
    },
    "hasMinimumDeviceForPhotosWeb": True,
    "iCDPEnabled": False,
    "webservices": {
        "reminders": {
            "url": "https://p31-remindersws.icloud.com:443",
            "status": "active",
        },
        "notes": {"url": "https://p38-notesws.icloud.com:443", "status": "active"},
        "mail": {"url": "https://p38-mailws.icloud.com:443", "status": "active"},
        "ckdatabasews": {
            "pcsRequired": True,
            "url": "https://p31-ckdatabasews.icloud.com:443",
            "status": "active",
        },
        "photosupload": {
            "pcsRequired": True,
            "url": ICLOUD_UPLOAD_PHOTOS_WS_URL,
            "status": "active",
        },
        "photos": {
            "pcsRequired": True,
            "uploadUrl": ICLOUD_UPLOAD_PHOTOS_WS_URL,
            "url": "https://p31-photosws.icloud.com:443",
            "status": "active",
        },
        "drivews": {
            "pcsRequired": True,
            "url": "https://p31-drivews.icloud.com:443",
            "status": "active",
        },
        "uploadimagews": {
            "url": "https://p31-uploadimagews.icloud.com:443",
            "status": "active",
        },
        "schoolwork": {},
        "cksharews": {"url": "https://p31-ckshare.icloud.com:443", "status": "active"},
        "findme": {"url": "https://p31-fmipweb.icloud.com:443", "status": "active"},
        "premiummailsettings": {
            "url": "https://p42-maildomainws.icloud.com:443",
            "status": "active",
        },
        "ckdeviceservice": {"url": "https://p31-ckdevice.icloud.com:443"},
        "iworkthumbnailws": {
            "url": "https://p31-iworkthumbnailws.icloud.com:443",
            "status": "active",
        },
        "calendar": {
            "url": "https://p31-calendarws.icloud.com:443",
            "status": "active",
        },
        "docws": {
            "pcsRequired": True,
            "url": "https://p31-docws.icloud.com:443",
            "status": "active",
        },
        "settings": {
            "url": "https://p31-settingsws.icloud.com:443",
            "status": "active",
        },
        "ubiquity": {
            "url": "https://p31-ubiquityws.icloud.com:443",
            "status": "active",
        },
        "streams": {"url": "https://p31-streams.icloud.com:443", "status": "active"},
        "keyvalue": {
            "url": "https://p31-keyvalueservice.icloud.com:443",
            "status": "active",
        },
        "archivews": {
            "url": "https://p31-archivews.icloud.com:443",
            "status": "active",
        },
        "push": {"url": "https://p31-pushws.icloud.com:443", "status": "active"},
        "iwmb": {"url": "https://p31-iwmb.icloud.com:443", "status": "active"},
        "iworkexportws": {
            "url": "https://p31-iworkexportws.icloud.com:443",
            "status": "active",
        },
        "geows": {"url": "https://p31-geows.icloud.com:443", "status": "active"},
        "account": {
            "iCloudEnv": {"shortId": "p", "vipSuffix": "prod"},
            "url": "https://p31-setup.icloud.com:443",
            "status": "active",
        },
        "fmf": {"url": "https://p31-fmfweb.icloud.com:443", "status": "active"},
        "contacts": {
            "url": "https://p31-contactsws.icloud.com:443",
            "status": "active",
        },
    },
    "pcsEnabled": True,
    "configBag": {
        "urls": {
            "accountCreateUI": ICLOUD_WIDGET_ACCOUNT_URL + WIDGET_KEY + "#!create",
            "accountLoginUI": f"{ICLOUD_AUTH_URL}/signin?widgetKey={WIDGET_KEY}",
            "accountLogin": "https://setup.icloud.com/setup/ws/1/accountLogin",
            "accountRepairUI": ICLOUD_WIDGET_ACCOUNT_URL + WIDGET_KEY + "#!repair",
            "downloadICloudTerms": "https://setup.icloud.com/setup/ws/1/downloadLiteTerms",
            "repairDone": "https://setup.icloud.com/setup/ws/1/repairDone",
            "accountAuthorizeUI": f"{ICLOUD_AUTH_URL}/authorize/signin?client_id={WIDGET_KEY}",
            "vettingUrlForEmail": "https://id.apple.com/IDMSEmailVetting/vetShareEmail",
            "accountCreate": "https://setup.icloud.com/setup/ws/1/createLiteAccount",
            "getICloudTerms": "https://setup.icloud.com/setup/ws/1/getTerms",
            "vettingUrlForPhone": "https://id.apple.com/IDMSEmailVetting/vetSharePhone",
        },
        "accountCreateEnabled": "true",
    },
    "hsaTrustedBrowser": False,
    "appsOrder": [
        "mail",
        "contacts",
        "calendar",
        "photos",
        "iclouddrive",
        "notes3",
        "reminders",
        "pages",
        "numbers",
        "keynote",
        "newspublisher",
        "fmf",
        "find",
        "settings",
    ],
    "version": 2,
    "isExtendedLogin": True,
    "pcsServiceIdentitiesIncluded": False,
    "hsaChallengeRequired": True,
    "ICDRSCapableDeviceCount": 1,
    "requestInfo": {"country": "FR", "timeZone": "GMT+1", "region": "IDF"},
    "pcsDeleted": False,
    "iCloudInfo": {"SafariBookmarksHasMigratedToCloudKit": True},
    "apps": {
        "calendar": {},
        "reminders": {},
        "keynote": {"isQualifiedForBeta": True},
        "settings": {"canLaunchWithOneFactor": True},
        "mail": {},
        "numbers": {"isQualifiedForBeta": True},
        "photos": {},
        "pages": {"isQualifiedForBeta": True},
        "notes3": {},
        "find": {"canLaunchWithOneFactor": True},
        "iclouddrive": {},
        "newspublisher": {"isHidden": True},
        "fmf": {},
        "contacts": {},
    },
}

TRUSTED_DEVICE_1: dict = {
    "deviceType": "SMS",
    "areaCode": "",
    "phoneNumber": "*******58",
    "deviceId": "1",
}
TRUSTED_DEVICES: dict = {"devices": [TRUSTED_DEVICE_1]}

VERIFICATION_CODE_OK: dict = {"success": True}
VERIFICATION_CODE_KO: dict = {"success": False}



"""Reminders service."""

import time
import uuid
from datetime import datetime
from typing import Any

from tzlocal import get_localzone_name

from pyicloud.services.base import BaseService
from pyicloud.session import PyiCloudSession


class RemindersService(BaseService):
    """The 'Reminders' iCloud service."""

    def __init__(
        self, service_root: str, session: PyiCloudSession, params: dict[str, Any]
    ) -> None:
        super().__init__(service_root, session, params)

        self.lists = {}
        self.collections = {}

        self.refresh()

    def refresh(self) -> None:
        """Refresh data."""
        params_reminders = dict(self.params)
        params_reminders.update(
            {
                "clientVersion": "4.0",
                "lang": "en-us",
                "usertz": get_localzone_name(),
            }
        )

        # Open reminders
        req = self.session.get(
            f"{self.service_root}/rd/startup", params=params_reminders
        )

        data = req.json()

        self.lists = {}
        self.collections = {}
        for collection in data["Collections"]:
            temp = []
            self.collections[collection["title"]] = {
                "guid": collection["guid"],
                "ctag": collection["ctag"],
            }
            for reminder in data["Reminders"]:
                if reminder["pGuid"] != collection["guid"]:
                    continue

                if reminder.get("dueDate"):
                    due = datetime(
                        reminder["dueDate"][0],
                        reminder["dueDate"][1],
                        reminder["dueDate"][2],
                        reminder["dueDate"][3],
                        reminder["dueDate"][4],
                        reminder["dueDate"][5],
                    )
                else:
                    due = None

                temp.append(
                    {
                        "title": reminder["title"],
                        "desc": reminder.get("description"),
                        "due": due,
                    }
                )
            self.lists[collection["title"]] = temp

    def post(self, title, description="", collection=None, due_date=None):
        """Adds a new reminder."""
        pguid = "tasks"
        if collection and collection in self.collections:
            pguid = self.collections[collection]["guid"]

        params_reminders = dict(self.params)
        params_reminders.update(
            {"clientVersion": "4.0", "lang": "en-us", "usertz": get_localzone_name()}
        )

        due_dates = None
        if due_date:
            due_dates = [
                int(f"{due_date.year}{due_date.month:02}{due_date.day:02}"),
                due_date.year,
                due_date.month,
                due_date.day,
                due_date.hour,
                due_date.minute,
            ]

        req = self.session.post(
            f"{self.service_root}/rd/reminders/tasks",
            json={
                "Reminders": {
                    "title": title,
                    "description": description,
                    "pGuid": pguid,
                    "etag": None,
                    "order": None,
                    "priority": 0,
                    "recurrence": None,
                    "alarms": [],
                    "startDate": None,
                    "startDateTz": None,
                    "startDateIsAllDay": False,
                    "completedDate": None,
                    "dueDate": due_dates,
                    "dueDateIsAllDay": False,
                    "lastModifiedDate": None,
                    "createdDate": None,
                    "isFamily": None,
                    "createdDateExtended": int(time.time() * 1000),
                    "guid": str(uuid.uuid4()),
                },
                "ClientState": {"Collections": list(self.collections.values())},
            },
            params=params_reminders,
        )
        return req.ok



"""Pytest configuration file for the pyicloud package."""
# pylint: disable=redefined-outer-name,protected-access

import os
import secrets
from typing import Any
from unittest.mock import MagicMock, mock_open, patch

import pytest
from requests.cookies import RequestsCookieJar

from pyicloud.base import PyiCloudService
from pyicloud.services.contacts import ContactsService
from pyicloud.services.drive import COOKIE_APPLE_WEBAUTH_VALIDATE
from pyicloud.services.hidemyemail import HideMyEmailService
from pyicloud.services.photos import BasePhotoLibrary, PhotoAsset
from pyicloud.session import PyiCloudSession
from tests import PyiCloudSessionMock
from tests.const import LOGIN_WORKING

BUILTINS_OPEN: str = "builtins.open"
EXAMPLE_DOMAIN: str = "https://example.com"


# pylint: disable=protected-access
# pylint: disable=redefined-outer-name


class FileSystemAccessError(Exception):
    """Raised when a test tries to access the file system."""


@pytest.fixture(autouse=True, scope="function")
def mock_file_open_write_fixture():
    """Mock the open function to prevent file system access."""
    # Dictionary to store written data
    written_data: dict[str, Any] = {}

    def mock_file_open(filepath: str, mode="r", **_):
        """Mock file open function."""

        if "w" in mode or "a" in mode:
            # Writing or appending mode
            def mock_write(content):
                if filepath not in written_data:
                    written_data[filepath] = ""
                if "a" in mode:  # Append mode
                    written_data[filepath] += content
                else:  # Write mode
                    written_data[filepath] = content

            mock_file = mock_open().return_value
            mock_file.write = mock_write
            return mock_file
        if "r" in mode:
            raise FileNotFoundError(f"No such file or directory: '{filepath}'")

        raise ValueError(f"Unsupported mode: {mode}")

    # Attach the written_data dictionary to the mock for access in tests
    mock_file_open.written_data = written_data  # type: ignore
    return mock_file_open


@pytest.fixture(autouse=True, scope="function")
def mock_mkdir():
    """Mock the mkdir function to prevent file system access."""
    mkdir = os.mkdir

    def my_mkdir(path, *args, **kwargs):
        if "python-test-results" not in path:
            raise FileSystemAccessError(
                f"You should not be creating directories in tests. {path}"
            )
        return mkdir(path, *args, **kwargs)

    with patch("os.mkdir", my_mkdir) as mkdir_mock:
        yield mkdir_mock


@pytest.fixture(autouse=True, scope="session")
def mock_open_fixture():
    """Mock the open function to prevent file system access."""
    builtins_open = open

    def my_open(path, *args, **kwargs):
        if "python-test-results" not in path:
            raise FileSystemAccessError(
                f"You should not be opening files in tests. {path}"
            )
        return builtins_open(path, *args, **kwargs)

    with patch(BUILTINS_OPEN, my_open) as open_mock:
        yield open_mock


@pytest.fixture(autouse=True, scope="session")
def mock_os_open_fixture():
    """Mock the open function to prevent file system access."""
    builtins_open = os.open

    def my_open(path, *args, **kwargs):
        if "python-test-results" not in path:
            raise FileSystemAccessError(
                f"You should not be opening files in tests. {path}"
            )
        return builtins_open(path, *args, **kwargs)

    with patch("os.open", my_open) as open_mock:
        yield open_mock


@pytest.fixture
def pyicloud_service() -> PyiCloudService:
    """Create a PyiCloudService instance with mocked authenticate method."""
    with (
        patch("pyicloud.PyiCloudService.authenticate") as mock_authenticate,
        patch(
            "pyicloud.PyiCloudService._setup_cookie_directory"
        ) as mock_setup_cookie_directory,
        patch(BUILTINS_OPEN, new_callable=mock_open),
    ):
        # Mock the authenticate method during initialization
        mock_authenticate.return_value = None
        mock_setup_cookie_directory.return_value = "/tmp/pyicloud/cookies"
        service = PyiCloudService("test@example.com", secrets.token_hex(32))
        return service


@pytest.fixture
def pyicloud_service_working(pyicloud_service: PyiCloudService) -> PyiCloudService:
    """Set the service to a working state."""
    pyicloud_service.data = LOGIN_WORKING
    pyicloud_service._webservices = LOGIN_WORKING["webservices"]
    with patch(BUILTINS_OPEN, new_callable=mock_open):
        pyicloud_service._session = PyiCloudSessionMock(
            pyicloud_service,
            "",
            cookie_directory="",
        )
        pyicloud_service.session._data = {"session_token": "valid_token"}
        check_pcs_consent = MagicMock(
            return_value={
                "isICDRSDisabled": False,
                "isDeviceConsentedForPCS": True,
            }
        )
        pyicloud_service._check_pcs_consent = check_pcs_consent

    return pyicloud_service


@pytest.fixture
def pyicloud_session(pyicloud_service_working: PyiCloudService) -> PyiCloudSession:
    """Mock the PyiCloudSession class."""
    pyicloud_service_working.session.cookies = MagicMock()
    return pyicloud_service_working.session


@pytest.fixture
def mock_session() -> MagicMock:
    """Fixture to create a mock PyiCloudSession."""
    return MagicMock(spec=PyiCloudSession)


@pytest.fixture
def contacts_service(mock_session: MagicMock) -> ContactsService:
    """Fixture to create a ContactsService instance."""
    return ContactsService(
        service_root=EXAMPLE_DOMAIN,
        session=mock_session,
        params={"test_param": "value"},
    )


@pytest.fixture
def mock_photos_service() -> MagicMock:
    """Fixture for mocking PhotosService."""
    service = MagicMock()
    service.service_endpoint = EXAMPLE_DOMAIN
    service.params = {"dsid": "12345"}
    service.session = MagicMock()
    return service


@pytest.fixture
def mock_photo_library(mock_photos_service: MagicMock) -> BasePhotoLibrary:
    """Fixture for mocking PhotoLibrary."""
    return BasePhotoLibrary(
        service=mock_photos_service,
        asset_type=PhotoAsset,
    )


@pytest.fixture
def hidemyemail_service(mock_session: MagicMock) -> HideMyEmailService:
    """Fixture for initializing HideMyEmailService."""
    return HideMyEmailService(EXAMPLE_DOMAIN, mock_session, {"dsid": "12345"})


@pytest.fixture
def mock_service_with_cookies(
    pyicloud_service_working: PyiCloudService,
) -> PyiCloudService:
    """Fixture to create a mock PyiCloudService with cookies."""
    jar = RequestsCookieJar()
    jar.set(COOKIE_APPLE_WEBAUTH_VALIDATE, "t=768y9u", domain="icloud.com", path="/")

    # Attach a real CookieJar so code that calls `.cookies.get()` keeps working.
    pyicloud_service_working.session.cookies = jar

    return pyicloud_service_working


@pytest.fixture(autouse=True, scope="session")
def mock_thread():
    """Mock threading.Thread to prevent actual thread creation during tests."""
    with patch("threading.Thread") as mock_thread_class:
        yield mock_thread_class



"""Services."""

from pyicloud.services.account import AccountService
from pyicloud.services.calendar import CalendarService
from pyicloud.services.contacts import ContactsService
from pyicloud.services.drive import DriveService
from pyicloud.services.findmyiphone import AppleDevice, FindMyiPhoneServiceManager
from pyicloud.services.hidemyemail import HideMyEmailService
from pyicloud.services.photos import PhotosService
from pyicloud.services.reminders import RemindersService
from pyicloud.services.ubiquity import UbiquityService

__all__: list[str] = [
    "AppleDevice",
    "AccountService",
    "CalendarService",
    "ContactsService",
    "DriveService",
    "FindMyiPhoneServiceManager",
    "HideMyEmailService",
    "PhotosService",
    "RemindersService",
    "UbiquityService",
]



"""
Tests for the utils module.
"""

import pytest

from pyicloud.utils import camelcase_to_underscore


@pytest.mark.parametrize(
    "camel_str,expected",
    [
        ("startDate", "start_date"),
        ("localStartDate", "local_start_date"),
        ("hasAttachments", "has_attachments"),
        ("simple", "simple"),
        ("CamelCase", "camel_case"),
        ("already_snake_case", "already_snake_case"),
        ("", ""),
        ("A", "a"),
        ("TestABC", "test_a_b_c"),
        ("testABC", "test_a_b_c"),
        ("testA", "test_a"),
        ("TestA", "test_a"),
        ("test", "test"),
        ("Test", "test"),
        ("testID", "test_i_d"),
        ("IDTest", "i_d_test"),
    ],
)
def test_camelcase_to_underscore(camel_str, expected):
    """Test the camelcase_to_underscore function."""
    assert camelcase_to_underscore(camel_str) == expected



"""Calendar service."""

import time
from calendar import monthrange
from dataclasses import asdict, dataclass, field, fields
from datetime import datetime, timedelta
from random import randint
from typing import Any, List, Literal, Optional, TypeVar, Union, cast, overload
from uuid import uuid4

from requests import Response
from tzlocal import get_localzone_name

from pyicloud.services.base import BaseService
from pyicloud.session import PyiCloudSession
from pyicloud.utils import camelcase_to_underscore

T = TypeVar("T")


# Calendar service constants
class DateFormats:
    """Date format constants."""

    API_DATE = "%Y-%m-%d"
    APPLE_DATE = "%Y%m%d"


class CalendarDefaults:
    """Default values for calendars."""

    TITLE = "Untitled"
    SYMBOLIC_COLOR = "__custom__"
    SUPPORTED_TYPE = "Event"
    OBJECT_TYPE = "personal"
    ORDER = 7
    SHARE_TITLE = ""
    SHARED_URL = ""
    COLOR = ""


class InviteeDefaults:
    """Default values for event invitees."""

    ROLE = "REQ-PARTICIPANT"
    STATUS = "NEEDS-ACTION"


class AlarmDefaults:
    """Default values for event alarms."""

    MESSAGE_TYPE = "message"
    IS_LOCATION_BASED = False


@dataclass
class AlarmMeasurement:
    """
    Represents the timing measurement for an alarm.
    """

    before: bool = True
    weeks: int = 0
    days: int = 0
    hours: int = 0
    minutes: int = 0
    seconds: int = 0


@dataclass
class AppleAlarm:
    """
    Represents an alarm in Apple's Calendar API format.

    This is used in the main payload's Alarm array.
    """

    # pylint: disable=invalid-name

    guid: str
    pGuid: str
    messageType: str = AlarmDefaults.MESSAGE_TYPE
    isLocationBased: bool = AlarmDefaults.IS_LOCATION_BASED
    measurement: AlarmMeasurement = field(default_factory=AlarmMeasurement)


@dataclass
class AppleDateFormat:
    """
    Apple's 7-element date array format.

    Format: [YYYYMMDD, YYYY, MM, DD, HH, MM, minutes_from_midnight]
    The date_string uses YYYYMMDD format.
    """

    date_string: str
    year: int
    month: int
    day: int
    hour: int
    minute: int
    minutes_from_midnight: int

    @classmethod
    def from_datetime(cls, dt: datetime, is_start: bool = True) -> "AppleDateFormat":
        """Convert Python datetime to Apple's date format."""
        if is_start:
            minutes_calc = dt.hour * 60 + dt.minute
        else:
            minutes_calc = (24 - dt.hour) * 60 + (60 - dt.minute)

        return cls(
            date_string=dt.strftime(DateFormats.APPLE_DATE),
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=dt.hour,
            minute=dt.minute,
            minutes_from_midnight=minutes_calc,
        )

    def to_list(self) -> list[int]:
        """Convert to Apple's expected list format."""
        return [
            self.date_string,
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.minutes_from_midnight,
        ]


@dataclass
class AppleEventInvitee:
    """
    Represents an invitee within the Event object in Apple's Calendar API.

    This is the simpler invitee structure used inside the Event.invitees array.
    """

    # pylint: disable=invalid-name

    email: str
    role: str = InviteeDefaults.ROLE
    inviteeStatus: str = InviteeDefaults.STATUS


@dataclass
class ApplePayloadInvitee:
    """
    Represents an invitee in the main payload's Invitee array in Apple's Calendar API.

    This is the more detailed invitee structure used in the top-level Invitee array.
    """

    # pylint: disable=invalid-name

    guid: str
    pGuid: str
    role: str = InviteeDefaults.ROLE
    isOrganizer: bool = False
    email: str = ""
    inviteeStatus: str = InviteeDefaults.STATUS
    commonName: str = ""
    isMe: bool = False


@dataclass
class AppleCalendarEvent:
    """
    Represents an event in Apple's Calendar API format.

    This dataclass exactly mirrors the structure expected by Apple's API,
    using the correct camelCase field names and types.
    """

    # pylint: disable=invalid-name

    title: str
    tz: str
    icon: int
    duration: int
    allDay: bool
    pGuid: str
    guid: str

    startDate: List[int]
    endDate: List[int]
    localStartDate: List[int]
    localEndDate: List[int]
    createdDate: List[int]
    lastModifiedDate: List[int]

    extendedDetailsAreIncluded: bool
    recurrenceException: bool
    recurrenceMaster: bool
    hasAttachments: bool
    readOnly: bool = False
    transparent: bool = False
    birthdayIsYearlessBday: bool = False
    birthdayShowAsCompany: bool = False
    shouldShowJunkUIWhenAppropriate: bool = False

    location: str = ""
    url: str = ""
    description: str = ""
    etag: str = ""

    alarms: List[str] = field(default_factory=list)
    attachments: List[Any] = field(default_factory=list)
    invitees: List[str] = field(default_factory=list)

    changeRecurring: Optional[str] = None


@dataclass
class EventObject:
    """
    An EventObject represents an event in the Apple Calendar.
    """

    pguid: str
    title: str = "New Event"
    start_date: datetime = field(default_factory=datetime.today)
    end_date: datetime = field(
        default_factory=lambda: datetime.today() + timedelta(minutes=60)
    )
    local_start_date: Optional[datetime] = None
    local_end_date: Optional[datetime] = None
    duration: int = field(init=False)
    icon: int = 0
    change_recurring: Optional[str] = None
    tz: str = ""
    guid: str = ""
    location: str = ""
    extended_details_are_included: bool = True
    recurrence_exception: bool = False
    recurrence_master: bool = False
    has_attachments: bool = False
    all_day: bool = False
    is_junk: bool = False
    etag: Optional[str] = None

    invitees: List[str] = field(init=False, default_factory=list)
    alarms: List[str] = field(init=False, default_factory=list)
    _alarm_metadata: dict[str, AlarmMeasurement] = field(
        init=False, default_factory=dict
    )

    def __post_init__(self) -> None:
        # Validation: check required fields and data integrity
        if not self.pguid.strip():
            raise ValueError("pguid cannot be empty")

        if self.start_date >= self.end_date:
            raise ValueError(
                f"start_date ({self.start_date}) must be before end_date ({self.end_date})"
            )

        # Initialize optional dates
        if not self.local_start_date:
            self.local_start_date = self.start_date
        if not self.local_end_date:
            self.local_end_date = self.end_date

        # Generate GUID if not provided
        if not self.guid:
            self.guid = str(uuid4()).upper()

        # Set timezone if not provided
        if not self.tz:
            self.tz = get_localzone_name()

        # Calculate duration (should now always be positive due to validation)
        self.duration = int(
            (self.end_date.timestamp() - self.start_date.timestamp()) / 60
        )

    def to_apple_event(self) -> AppleCalendarEvent:
        """
        Convert this EventObject to Apple's API format.

        This method handles the conversion from our internal representation
        to the exact structure expected by Apple's Calendar API.
        """
        # Convert Python datetime to Apple's date format
        start_date_list = self.dt_to_list(self.start_date)
        end_date_list = self.dt_to_list(self.end_date, False)
        local_start_list = (
            self.dt_to_list(self.local_start_date)
            if self.local_start_date
            else start_date_list
        )
        local_end_list = (
            self.dt_to_list(self.local_end_date, False)
            if self.local_end_date
            else end_date_list
        )

        current_timestamp = time.time()
        current_dt = datetime.fromtimestamp(current_timestamp)
        created_date_list = self.dt_to_list(current_dt)
        last_modified_list = self.dt_to_list(current_dt)

        invitees_list: List[str] = []
        if self.invitees:
            invitees_list = self.invitees

        alarms_list: List[str] = []
        if self.alarms:
            alarms_list = self.alarms

        return AppleCalendarEvent(
            title=self.title,
            tz=self.tz,
            icon=self.icon,
            duration=self.duration,
            allDay=self.all_day,
            pGuid=self.pguid,
            guid=self.guid,
            startDate=start_date_list,
            endDate=end_date_list,
            localStartDate=local_start_list,
            localEndDate=local_end_list,
            createdDate=created_date_list,
            lastModifiedDate=last_modified_list,
            extendedDetailsAreIncluded=self.extended_details_are_included,
            recurrenceException=self.recurrence_exception,
            recurrenceMaster=self.recurrence_master,
            hasAttachments=self.has_attachments,
            shouldShowJunkUIWhenAppropriate=self.is_junk,
            location=self.location,
            etag=self.etag or "",
            alarms=alarms_list,
            invitees=invitees_list,
            changeRecurring=self.change_recurring,
        )

    @property
    def request_data(self) -> dict[str, Any]:
        """Returns the event data in the format required by Apple's calendar."""
        apple_event = self.to_apple_event()
        event_dict = asdict(apple_event)

        data: dict[str, Any] = {
            "Event": event_dict,
            "Invitee": [],
            "Alarm": [],
            "ClientState": {
                "Collection": [{"guid": self.pguid, "ctag": None}],
            },
        }

        if self.invitees:
            payload_invitees = [
                ApplePayloadInvitee(
                    guid=email_guid,
                    pGuid=self.guid,
                    role=InviteeDefaults.ROLE,
                    isOrganizer=False,
                    email=email_guid.split(":")[-1],
                    inviteeStatus=InviteeDefaults.STATUS,
                    commonName="",
                    isMe=False,
                )
                for email_guid in self.invitees
            ]
            data["Invitee"] = [asdict(invitee) for invitee in payload_invitees]

        if self.alarms:
            payload_alarms = [
                AppleAlarm(
                    guid=alarm_guid,
                    pGuid=self.guid,
                    messageType=AlarmDefaults.MESSAGE_TYPE,
                    isLocationBased=AlarmDefaults.IS_LOCATION_BASED,
                    measurement=self._alarm_metadata.get(
                        alarm_guid, AlarmMeasurement()
                    ),
                )
                for alarm_guid in self.alarms
            ]
            data["Alarm"] = [asdict(alarm) for alarm in payload_alarms]

        return data

    def dt_to_list(self, dt: datetime, start: bool = True) -> list:
        """
        Converts python datetime object into a list format used
        by Apple's calendar.
        """
        apple_date = AppleDateFormat.from_datetime(dt, is_start=start)
        return apple_date.to_list()

    def add_invitees(self, _invitees: Optional[list] = None) -> None:
        """
        Adds a list of emails to invitees in the correct format
        """
        if _invitees:
            self.invitees += [f"{self.guid}:{email}" for email in _invitees]

    def add_alarm_at_time(self) -> str:
        """
        Adds an alarm at the time of the event.
        Returns the alarm GUID for reference.
        """
        alarm_guid = str(uuid4())
        alarm_full_guid = f"{self.guid}:{alarm_guid}"
        self.alarms.append(alarm_full_guid)

        self._alarm_metadata[alarm_full_guid] = AlarmMeasurement(
            before=False, weeks=0, days=0, hours=0, minutes=0, seconds=0
        )
        return alarm_guid

    def add_alarm_before(
        self, minutes: int = 0, hours: int = 0, days: int = 0, weeks: int = 0
    ) -> str:
        """
        Adds an alarm before the event.

        Args:
            minutes: Minutes before event
            hours: Hours before event
            days: Days before event
            weeks: Weeks before event

        Returns:
            The alarm GUID for reference.
        """
        alarm_guid = str(uuid4())
        alarm_full_guid = f"{self.guid}:{alarm_guid}"
        self.alarms.append(alarm_full_guid)

        self._alarm_metadata[alarm_full_guid] = AlarmMeasurement(
            before=True, weeks=weeks, days=days, hours=hours, minutes=minutes, seconds=0
        )
        return alarm_guid

    def get(self, var: str):
        """Get a variable"""
        return getattr(self, var, None)


@dataclass
class CalendarObject:
    """
    A CalendarObject represents a calendar in the Apple Calendar.
    """

    title: str = CalendarDefaults.TITLE
    guid: str = ""
    share_type: Optional[str] = None
    symbolic_color: str = CalendarDefaults.SYMBOLIC_COLOR
    supported_type: str = CalendarDefaults.SUPPORTED_TYPE
    object_type: str = CalendarDefaults.OBJECT_TYPE
    share_title: str = CalendarDefaults.SHARE_TITLE
    shared_url: str = CalendarDefaults.SHARED_URL
    color: str = CalendarDefaults.COLOR
    order: int = CalendarDefaults.ORDER
    extended_details_are_included: bool = True
    read_only: bool = False
    enabled: bool = True
    ignore_event_updates: Optional[str] = None
    email_notification: Optional[str] = None
    last_modified_date: Optional[str] = None
    me_as_participant: Optional[str] = None
    pre_published_url: Optional[str] = None
    participants: Optional[str] = None
    defer_loading: Optional[str] = None
    published_url: Optional[str] = None
    remove_alarms: Optional[str] = None
    ignore_alarms: Optional[str] = None
    description: Optional[str] = None
    remove_todos: Optional[str] = None
    is_default: Optional[bool] = None
    is_family: Optional[bool] = None
    etag: Optional[str] = None
    ctag: Optional[str] = None

    def __post_init__(self) -> None:
        if not self.guid:
            self.guid = str(uuid4()).upper()

        if not self.color:
            self.color = self.gen_random_color()

    def gen_random_color(self) -> str:
        """
        Creates a random rgbhex color.
        """
        return f"#{randint(0, 255):02x}{randint(0, 255):02x}{randint(0, 255):02x}"

    @property
    def request_data(self) -> dict[str, Any]:
        """Returns the calendar data in the format required by Apple's calendar."""
        data: dict[str, Any] = {
            "Collection": asdict(self),
            "ClientState": {
                "Collection": [],
                "fullState": False,
            },
        }
        return data


class CalendarService(BaseService):
    """
    The 'Calendar' iCloud service, connects to iCloud and returns events.
    """

    def __init__(
        self, service_root: str, session: PyiCloudSession, params: dict[str, Any]
    ) -> None:
        super().__init__(service_root, session, params)
        self._calendar_endpoint: str = f"{self.service_root}/ca"
        self._calendar_refresh_url: str = f"{self._calendar_endpoint}/events"
        self._calendar_event_detail_url: str = f"{self._calendar_endpoint}/eventdetail"
        self._calendar_collections_url: str = f"{self._calendar_endpoint}/collections"
        self._calendars_url: str = f"{self._calendar_endpoint}/allcollections"

    @property
    def default_params(self) -> dict[str, Any]:
        """Returns the default parameters for the calendar service."""
        today: datetime = datetime.today()
        _, days_in_month = monthrange(today.year, today.month)
        # monthrange returns: weekday of the first day of the month (0 -> Mon, 6 -> Sun) and
        # number of days in the month (Jan -> 31, Feb -> 28/29, etc.)
        from_dt = datetime(
            today.year, today.month, 1
        )  # Hardcoded to 1 so that startDate is always the first (1st) day of the month
        to_dt = datetime(today.year, today.month, days_in_month)
        params = dict(self.params)
        params.update(
            {
                "lang": "en-us",
                "usertz": get_localzone_name(),
                "startDate": from_dt.strftime(DateFormats.API_DATE),
                "endDate": to_dt.strftime(DateFormats.API_DATE),
            }
        )

        return params

    def obj_from_dict(self, obj: T, _dict) -> T:
        """Creates an object from a dictionary with proper field validation."""
        if hasattr(obj, "__dataclass_fields__"):
            valid_fields = {f.name for f in fields(obj)}

            special_mappings = {
                "pGuid": "pguid",
                "shouldShowJunkUIWhenAppropriate": "is_junk",
            }

            for api_key, value in _dict.items():
                field_name: str = special_mappings.get(
                    api_key, camelcase_to_underscore(api_key)
                )

                if field_name in valid_fields:
                    setattr(obj, field_name, value)
        else:
            for key, value in _dict.items():
                setattr(obj, key, value)

        return obj

    def get_ctag(self, guid: str) -> str:
        """Returns the ctag for a given calendar guid"""
        ctag: Optional[str] = None
        for cal in self.get_calendars(as_objs=False):
            if isinstance(cal, CalendarObject) and cal.guid == guid:
                ctag = cal.ctag
            elif isinstance(cal, dict) and cal.get("guid") == guid:
                ctag = cal.get("ctag")

            if ctag:
                return ctag
        raise ValueError("ctag not found.")

    def refresh_client(self, from_dt=None, to_dt=None) -> dict[str, Any]:
        """
        Refresh the Calendar service and return a fresh event payload.

        Date range semantics:
        - If both 'from_dt' and 'to_dt' are provided, they are respected as-is.
        - If exactly one bound is provided, the missing bound is anchored to the
          same month as the provided bound and expanded to the full month
          (1st..last day of that month).
        - If neither is provided, defaults to the current month.

        Notes:
        - Apple's Calendar API treats 'endDate' as inclusive; events occurring on
          the last day of the month (including all-day and 23:00->00:00 boundary
          events) are returned. See tests in tests/test_calendar.py.
        """
        today: datetime = datetime.today()
        # Anchor missing bound(s) to whichever bound is provided, else to 'today'
        anchor: datetime = from_dt or to_dt or today
        year, month = anchor.year, anchor.month
        _, days_in_month = monthrange(
            year, month
        )  # (weekday_of_first_day, days_in_month)
        # If either bound is missing, normalize to the full month of the anchor.
        # When both bounds are provided (e.g., day/week queries), respect them as-is.
        if from_dt is None or to_dt is None:
            from_dt = anchor.replace(day=1)
            to_dt = anchor.replace(day=days_in_month)
        params = dict(self.params)
        params.update(
            {
                "lang": "en-us",
                "usertz": get_localzone_name(),
                "startDate": from_dt.strftime(DateFormats.API_DATE),
                "endDate": to_dt.strftime(DateFormats.API_DATE),
                "dsid": self.session.service.data["dsInfo"]["dsid"],
            }
        )
        req: Response = self.session.get(self._calendar_refresh_url, params=params)
        return req.json()

    @overload
    def get_calendars(self) -> list[dict[str, Any]]: ...

    @overload
    def get_calendars(self, as_objs: Literal[False]) -> list[dict[str, Any]]: ...

    @overload
    def get_calendars(self, as_objs: Literal[True]) -> list[CalendarObject]: ...

    def get_calendars(
        self, as_objs: Union[Literal[True], Literal[False]] = False
    ) -> Union[list[dict[str, Any]], list[CalendarObject]]:
        """
        Retrieves calendars of this month.
        """
        params: dict[str, Any] = self.default_params
        req: Response = self.session.get(self._calendars_url, params=params)
        response = req.json()
        calendars: list[dict[str, Any]] = response["Collection"]

        if not as_objs and calendars:
            return calendars

        return [self.obj_from_dict(CalendarObject(), cal) for cal in calendars]

    def add_calendar(self, calendar: CalendarObject) -> dict[str, Any]:
        """
        Adds a Calendar to the apple calendar.
        """
        data: dict[str, Any] = calendar.request_data
        params: dict[str, Any] = self.default_params

        req: Response = self.session.post(
            f"{self._calendar_collections_url}/{calendar.guid}",
            params=params,
            json=data,
        )
        return req.json()

    def remove_calendar(self, cal_guid: str) -> dict[str, Any]:
        """
        Removes a Calendar from the apple calendar.
        """
        params: dict[str, Any] = self.default_params
        params["methodOverride"] = "DELETE"

        req: Response = self.session.post(
            f"{self._calendar_collections_url}/{cal_guid}", params=params, json={}
        )
        return req.json()

    @overload
    def get_events(
        self,
        from_dt: Optional[datetime] = None,
        to_dt: Optional[datetime] = None,
        period: str = "month",
    ) -> list[dict[str, Any]]: ...

    @overload
    def get_events(
        self,
        from_dt: Optional[datetime] = None,
        to_dt: Optional[datetime] = None,
        period: str = "month",
        as_objs: Literal[False] = False,
    ) -> list[dict[str, Any]]: ...

    @overload
    def get_events(
        self,
        from_dt: Optional[datetime] = None,
        to_dt: Optional[datetime] = None,
        period: str = "month",
        as_objs: Literal[True] = True,
    ) -> list[EventObject]: ...

    def get_events(
        self,
        from_dt: Optional[datetime] = None,
        to_dt: Optional[datetime] = None,
        period: str = "month",
        as_objs: bool = False,
    ) -> Union[list[dict[str, Any]], list[EventObject]]:
        """
        Retrieves events for a given date range, by default, this month.
        """
        today: datetime = datetime.today()
        if period != "month" and from_dt:
            today = datetime(from_dt.year, from_dt.month, from_dt.day)

        if period == "day":
            if not from_dt:
                from_dt = datetime(today.year, today.month, today.day)
            to_dt = from_dt + timedelta(days=1)
        elif period == "week":
            if not from_dt:
                from_dt = datetime(today.year, today.month, today.day) - timedelta(
                    days=today.weekday() + 1
                )
            to_dt = from_dt + timedelta(days=6)

        response: dict[str, Any] = self.refresh_client(from_dt, to_dt)
        events: list = response.get("Event", [])

        if as_objs and events:
            for idx, event in enumerate(events):
                pguid = event.get("pGuid", "")
                if not pguid:
                    raise ValueError(f"Event missing required pGuid field: {event}")
                events[idx] = self.obj_from_dict(EventObject(pguid), event)

        return events

    def get_event_detail(self, pguid, guid, as_obj: bool = False) -> EventObject:
        """
        Fetches a single event's details by specifying a pguid
        (a calendar) and a guid (an event's ID).
        """
        params = dict(self.params)
        params.update(
            {
                "lang": "en-us",
                "usertz": get_localzone_name(),
                "dsid": self.session.service.data["dsInfo"]["dsid"],
            }
        )
        url: str = f"{self._calendar_event_detail_url}/{pguid}/{guid}"
        req: Response = self.session.get(url, params=params)
        response = req.json()
        event = response["Event"][0]

        if as_obj and event:
            event: EventObject = cast(
                EventObject,
                self.obj_from_dict(EventObject(pguid=pguid), event),
            )

        return event

    def add_event(self, event: EventObject) -> dict[str, Any]:
        """
        Adds an Event to a calendar.
        """
        data = event.request_data
        data["ClientState"]["Collection"][0]["ctag"] = self.get_ctag(event.pguid)
        params = self.default_params

        req: Response = self.session.post(
            f"{self._calendar_refresh_url}/{event.pguid}/{event.guid}",
            params=params,
            json=data,
        )
        return req.json()

    def remove_event(self, event: EventObject) -> dict[str, Any]:
        """
        Removes an Event from a calendar. The calendar's guid corresponds to the EventObject's pGuid
        """
        data = event.request_data
        data["ClientState"]["Collection"][0]["ctag"] = self.get_ctag(event.pguid)
        data["Event"] = {}

        params: dict[str, Any] = self.default_params
        params["methodOverride"] = "DELETE"
        if not getattr(event, "etag", None):
            event.etag = self.get_event_detail(
                event.pguid, event.guid, as_obj=False
            ).get("etag")
        params["ifMatch"] = event.etag

        req: Response = self.session.post(
            f"{self._calendar_refresh_url}/{event.pguid}/{event.guid}",
            params=params,
            json=data,
        )
        return req.json()



"""Drive service tests."""
# pylint: disable=protected-access

from typing import Optional
from unittest.mock import ANY, Mock, patch

import pytest

from pyicloud import PyiCloudService
from pyicloud.const import CONTENT_TYPE, CONTENT_TYPE_TEXT
from pyicloud.exceptions import PyiCloudAPIResponseException
from pyicloud.services.drive import (
    CLOUD_DOCS_ZONE,
    NODE_TRASH,
    DriveNode,
    DriveService,
)


def test_root(pyicloud_service_working: PyiCloudService) -> None:
    """Test the root folder."""
    drive: DriveService = pyicloud_service_working.drive
    # root name is now extracted from drivewsid.
    assert drive.name == "root"
    assert drive.type == "folder"
    assert drive.size is None
    assert drive.date_changed is None
    assert drive.date_modified is None
    assert drive.date_last_open is None
    assert drive.dir() == ["Keynote", "Numbers", "Pages", "Preview", "pyiCloud"]


def test_trash(pyicloud_service_working: PyiCloudService) -> None:
    """Test the trash folder."""
    trash: DriveNode = pyicloud_service_working.drive.trash
    assert trash.name == NODE_TRASH
    assert trash.type == DriveNode.TYPE_TRASH
    assert trash.size is None
    assert trash.date_changed is None
    assert trash.date_modified is None
    assert trash.date_last_open is None
    assert trash.dir() == [
        "dead-file.download",
        "test_create_folder",
        "test_delete_forever_and_ever",
        "test_files_1",
        "test_random_uuid",
        "test12345",
    ]


def test_trash_recover(pyicloud_service_working: PyiCloudService) -> None:
    """Test recovering a file from the Trash."""
    trash_node = pyicloud_service_working.drive.trash["test_random_uuid"]
    assert trash_node is not None
    recover_result = trash_node.recover()
    recover_result_items = recover_result["items"][0]
    assert recover_result_items["status"] == "OK"
    assert recover_result_items["parentId"] == "FOLDER::com.apple.CloudDocs::root"
    assert recover_result_items["name"] == "test_random_uuid"


def test_trash_delete_forever(pyicloud_service_working: PyiCloudService) -> None:
    """Test permanently deleting a file from the Trash."""
    node = pyicloud_service_working.drive.trash["test_delete_forever_and_ever"]
    assert node is not None, "Expected a valid trash node before deleting forever."
    recover_result = node.delete_forever()
    recover_result_items = recover_result["items"][0]
    assert recover_result_items["status"] == "OK"
    assert (
        recover_result_items["parentId"]
        == "FOLDER::com.apple.CloudDocs::43D7C666-6E6E-4522-8999-0B519C3A1F4B"
    )
    assert recover_result_items["name"] == "test_delete_forever_and_ever"


def test_folder_app(pyicloud_service_working: PyiCloudService) -> None:
    """Test the /Preview folder."""
    folder: Optional[DriveNode] = pyicloud_service_working.drive["Preview"]
    assert folder
    assert folder.name == "Preview"
    assert folder.type == "app_library"
    assert folder.size is None
    assert folder.date_changed is None
    assert folder.date_modified is None
    assert folder.date_last_open is None
    with pytest.raises(KeyError, match="No items in folder, status: ID_INVALID"):
        folder.dir()


def test_folder_not_exists(pyicloud_service_working: PyiCloudService) -> None:
    """Test the /not_exists folder."""
    with pytest.raises(KeyError, match="No child named 'not_exists' exists"):
        _ = pyicloud_service_working.drive["not_exists"]


def test_folder(pyicloud_service_working: PyiCloudService) -> None:
    """Test the /pyiCloud folder."""
    folder: Optional[DriveNode] = pyicloud_service_working.drive["pyiCloud"]
    assert folder
    assert folder.name == "pyiCloud"
    assert folder.type == "folder"
    assert folder.size is None
    assert folder.date_changed is None
    assert folder.date_modified is None
    assert folder.date_last_open is None
    assert folder.dir() == ["Test"]


def test_subfolder(pyicloud_service_working: PyiCloudService) -> None:
    """Test the /pyiCloud/Test folder."""
    parent_folder: Optional[DriveNode] = pyicloud_service_working.drive["pyiCloud"]
    assert parent_folder is not None, "Expected to find 'pyiCloud' folder."
    folder: Optional[DriveNode] = parent_folder["Test"]
    assert folder
    assert folder.name == "Test"
    assert folder.type == "folder"
    assert folder.size is None
    assert folder.date_changed is None
    assert folder.date_modified is None
    assert folder.date_last_open is None
    assert folder.dir() == ["Document scanné 2.pdf", "Scanned document 1.pdf"]


def test_subfolder_file(pyicloud_service_working: PyiCloudService) -> None:
    """Test the /pyiCloud/Test/Scanned document 1.pdf file."""
    drive: Optional[DriveNode] = pyicloud_service_working.drive["pyiCloud"]
    assert drive
    folder: Optional[DriveNode] = drive["Test"]
    assert folder
    file_test: Optional[DriveNode] = folder["Scanned document 1.pdf"]
    assert file_test
    assert file_test.name == "Scanned document 1.pdf"
    assert file_test.type == "file"
    assert file_test.size == 21644358
    assert str(file_test.date_changed) == "2020-05-03 00:16:17"
    assert str(file_test.date_modified) == "2020-05-03 00:15:17"
    assert str(file_test.date_last_open) == "2020-05-03 00:24:25"
    with pytest.raises(NotADirectoryError):
        file_test.dir()


def test_file_open(pyicloud_service_working: PyiCloudService) -> None:
    """Test the /pyiCloud/Test/Scanned document 1.pdf file open."""
    drive: Optional[DriveNode] = pyicloud_service_working.drive["pyiCloud"]
    assert drive
    folder: Optional[DriveNode] = drive["Test"]
    assert folder
    file_test: Optional[DriveNode] = folder["Scanned document 1.pdf"]
    assert file_test
    with file_test.open(stream=True) as response:
        assert response.raw


def test_get_node_data(pyicloud_service_working: PyiCloudService) -> None:
    """Test retrieving node data."""
    drive: DriveService = pyicloud_service_working.drive
    mock_response = {"drivewsid": "test_id", "name": "Test Node"}
    with patch.object(
        drive.session, "post", return_value=Mock(ok=True, json=lambda: [mock_response])
    ) as mock_post:
        node_data = drive.get_node_data("test_id")
        assert node_data == mock_response
        mock_post.assert_called_once_with(
            drive.service_root + "/retrieveItemDetailsInFolders",
            params=drive.params,
            json=[{"drivewsid": "test_id", "partialData": False}],
        )


def test_get_file(pyicloud_service_working: PyiCloudService) -> None:
    """Test retrieving a file."""
    drive: DriveService = pyicloud_service_working.drive
    mock_response = {"data_token": {"url": "https://example.com/file"}}
    with patch.object(
        drive.session,
        "get",
        side_effect=[
            Mock(ok=True, json=lambda: mock_response),
            Mock(ok=True, content=b"file content"),
        ],
    ) as mock_get:
        file_response = drive.get_file("file_id")
        assert file_response.content == b"file content"
        mock_get.assert_any_call(
            drive._document_root + f"/ws/{CLOUD_DOCS_ZONE}/download/by_id",
            params={**drive.params, "document_id": "file_id"},
        )
        mock_get.assert_any_call("https://example.com/file", params=drive.params)


def test_create_folders(pyicloud_service_working: PyiCloudService) -> None:
    """Test creating a folder."""
    drive: DriveService = pyicloud_service_working.drive
    mock_response = {"folders": [{"name": "New Folder"}]}
    with patch.object(
        drive.session, "post", return_value=Mock(ok=True, json=lambda: mock_response)
    ) as mock_post:
        response = drive.create_folders("parent_id", "New Folder")
        assert response == mock_response
        mock_post.assert_called_once_with(
            drive.service_root + "/createFolders",
            params=drive.params,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json=ANY,
        )


def test_delete_items(pyicloud_service_working: PyiCloudService) -> None:
    """Test deleting an item."""
    drive: DriveService = pyicloud_service_working.drive
    mock_response = {"status": "OK"}
    with patch.object(
        drive.session, "post", return_value=Mock(ok=True, json=lambda: mock_response)
    ) as mock_post:
        response = drive.delete_items("node_id", "etag")
        assert response == mock_response
        mock_post.assert_called_once_with(
            drive.service_root + "/deleteItems",
            params=drive.params,
            json={
                "items": [
                    {
                        "drivewsid": "node_id",
                        "etag": "etag",
                        "clientId": drive.params["clientId"],
                    },
                ]
            },
        )


def test_rename_items(pyicloud_service_working: PyiCloudService) -> None:
    """Test renaming an item."""
    drive: DriveService = pyicloud_service_working.drive
    mock_response = {"status": "OK"}
    with patch.object(
        drive.session,
        "post",
        return_value=Mock(ok=True, json=lambda: mock_response),
    ) as mock_post:
        response = drive.rename_items("node_id", "etag", "New Name")
        assert response == mock_response
        mock_post.assert_called_once_with(
            drive.service_root + "/renameItems",
            params=drive.params,
            json={
                "items": [
                    {"drivewsid": "node_id", "etag": "etag", "name": "New Name"},
                ]
            },
        )


def test_move_items_to_trash(pyicloud_service_working: PyiCloudService) -> None:
    """Test moving an item to trash."""
    drive: DriveService = pyicloud_service_working.drive
    mock_response = {"status": "OK"}
    with patch.object(
        drive.session, "post", return_value=Mock(ok=True, json=lambda: mock_response)
    ) as mock_post:
        response = drive.move_items_to_trash("node_id", "etag")
        assert response == mock_response
        mock_post.assert_called_once_with(
            drive.service_root + "/moveItemsToTrash",
            params=drive.params,
            json={
                "items": [
                    {"drivewsid": "node_id", "etag": "etag", "clientId": "node_id"},
                ]
            },
        )


def test_recover_items_from_trash(pyicloud_service_working: PyiCloudService) -> None:
    """Test recovering an item from trash."""
    drive: DriveService = pyicloud_service_working.drive
    mock_response = {"status": "OK"}
    with patch.object(
        drive.session, "post", return_value=Mock(ok=True, json=lambda: mock_response)
    ) as mock_post:
        response = drive.recover_items_from_trash("node_id", "etag")
        assert response == mock_response
        mock_post.assert_called_once_with(
            drive.service_root + "/putBackItemsFromTrash",
            params=drive.params,
            json={
                "items": [
                    {"drivewsid": "node_id", "etag": "etag"},
                ]
            },
        )


def test_delete_forever_from_trash(pyicloud_service_working: PyiCloudService) -> None:
    """Test permanently deleting an item from trash."""
    drive: DriveService = pyicloud_service_working.drive
    mock_response = {"status": "OK"}
    with patch.object(
        drive.session, "post", return_value=Mock(ok=True, json=lambda: mock_response)
    ) as mock_post:
        response = drive.delete_forever_from_trash("node_id", "etag")
        assert response == mock_response
        mock_post.assert_called_once_with(
            drive.service_root + "/deleteItems",
            params=drive.params,
            json={
                "items": [
                    {"drivewsid": "node_id", "etag": "etag"},
                ]
            },
        )


def test_get_upload_contentws_url_success(
    mock_service_with_cookies: PyiCloudService,
) -> None:
    """Test successful retrieval of upload contentWS URL."""
    drive: DriveService = mock_service_with_cookies.drive
    mock_file = Mock()
    mock_file.name = "test_file.txt"
    mock_file.tell = Mock(side_effect=[0, 100, 0])  # Mock file size as 100 bytes

    mock_response = [
        {"document_id": "mock_document_id", "url": "https://example.com/upload"}
    ]
    with (
        patch.object(
            drive.session,
            "post",
            return_value=Mock(ok=True, json=lambda: mock_response),
        ) as mock_post,
        patch("mimetypes.guess_type", return_value=("text/plain", None)),
    ):
        document_id, url = drive._get_upload_contentws_url(mock_file)

        assert document_id == "mock_document_id"
        assert url == "https://example.com/upload"

        mock_post.assert_called_once_with(
            drive._document_root + f"/ws/{CLOUD_DOCS_ZONE}/upload/web",
            params=ANY,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json={
                "filename": "test_file.txt",
                "type": "FILE",
                "content_type": "text/plain",
                "size": 100,
            },
        )


def test_get_upload_contentws_url_no_content_type(
    mock_service_with_cookies: PyiCloudService,
) -> None:
    """Test retrieval of upload contentWS URL when content type is None."""
    drive: DriveService = mock_service_with_cookies.drive
    mock_file = Mock()
    mock_file.name = "test_file.unknown"
    mock_file.tell = Mock(side_effect=[0, 200, 0])  # Mock file size as 200 bytes

    mock_response = [
        {"document_id": "mock_document_id", "url": "https://example.com/upload"}
    ]
    with (
        patch.object(
            drive.session,
            "post",
            return_value=Mock(ok=True, json=lambda: mock_response),
        ) as mock_post,
        patch("mimetypes.guess_type", return_value=(None, None)),
    ):
        document_id, url = drive._get_upload_contentws_url(mock_file)

        assert document_id == "mock_document_id"
        assert url == "https://example.com/upload"

        mock_post.assert_called_once_with(
            drive._document_root + f"/ws/{CLOUD_DOCS_ZONE}/upload/web",
            params=ANY,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json={
                "filename": "test_file.unknown",
                "type": "FILE",
                "content_type": "",
                "size": 200,
            },
        )


def test_get_upload_contentws_url_error_response(
    mock_service_with_cookies: PyiCloudService,
) -> None:
    """Test retrieval of upload contentWS URL with an error response."""
    drive: DriveService = mock_service_with_cookies.drive
    mock_file = Mock()
    mock_file.name = "test_file.txt"
    mock_file.tell = Mock(side_effect=[0, 300, 0])  # Mock file size as 300 bytes

    with (
        patch.object(
            drive.session, "post", return_value=Mock(ok=False, reason="Bad Request")
        ) as mock_post,
        patch("mimetypes.guess_type", return_value=("text/plain", None)),
    ):
        with pytest.raises(PyiCloudAPIResponseException, match="Bad Request"):
            drive._get_upload_contentws_url(mock_file)

        mock_post.assert_called_once_with(
            drive._document_root + f"/ws/{CLOUD_DOCS_ZONE}/upload/web",
            params=ANY,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json={
                "filename": "test_file.txt",
                "type": "FILE",
                "content_type": "text/plain",
                "size": 300,
            },
        )


def test_get_upload_contentws_url_invalid_response_format(
    mock_service_with_cookies: PyiCloudService,
) -> None:
    """Test retrieval of upload contentWS URL with an invalid response format."""
    drive: DriveService = mock_service_with_cookies.drive
    mock_file = Mock()
    mock_file.name = "test_file.txt"
    mock_file.tell = Mock(side_effect=[0, 400, 0])  # Mock file size as 400 bytes

    mock_response = []  # Invalid response format
    with (
        patch.object(
            drive.session,
            "post",
            return_value=Mock(ok=True, json=lambda: mock_response),
        ) as mock_post,
        patch("mimetypes.guess_type", return_value=("text/plain", None)),
    ):
        with pytest.raises(IndexError):
            drive._get_upload_contentws_url(mock_file)

        mock_post.assert_called_once_with(
            drive._document_root + f"/ws/{CLOUD_DOCS_ZONE}/upload/web",
            params=ANY,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json={
                "filename": "test_file.txt",
                "type": "FILE",
                "content_type": "text/plain",
                "size": 400,
            },
        )


def test_send_file_success(mock_service_with_cookies: PyiCloudService) -> None:
    """Test successfully sending a file to iCloud Drive."""
    drive: DriveService = mock_service_with_cookies.drive

    mock_file = Mock()
    mock_file.name = "test_file.txt"
    mock_file.tell = Mock(side_effect=[0, 100, 0])  # Mock file size as 100 bytes

    mock_upload_url_response = [
        {"document_id": "mock_document_id", "url": "https://example.com/upload"}
    ]
    mock_upload_response = {
        "singleFile": {
            "fileChecksum": "mock_checksum",
            "wrappingKey": "mock_key",
            "referenceChecksum": "mock_reference",
            "size": 100,
        }
    }
    mock_update_response = {"status": "OK"}

    with (
        patch.object(
            drive.session,
            "post",
            side_effect=[
                Mock(
                    ok=True, json=lambda: mock_upload_url_response
                ),  # _get_upload_contentws_url
                Mock(ok=True, json=lambda: mock_upload_response),  # Upload file
                Mock(ok=True, json=lambda: mock_update_response),  # _update_contentws
            ],
        ) as mock_post,
        patch("mimetypes.guess_type", return_value=("text/plain", None)),
    ):
        drive.send_file("mock_folder_id", mock_file)

        # Assert _get_upload_contentws_url call
        mock_post.assert_any_call(
            drive._document_root + f"/ws/{CLOUD_DOCS_ZONE}/upload/web",
            params=ANY,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json={
                "filename": "test_file.txt",
                "type": "FILE",
                "content_type": "text/plain",
                "size": 100,
            },
        )

        # Assert file upload call
        mock_post.assert_any_call(
            "https://example.com/upload",
            files={"test_file.txt": mock_file},
        )

        # Assert _update_contentws call
        mock_post.assert_any_call(
            drive._document_root + f"/ws/{CLOUD_DOCS_ZONE}/update/documents",
            params=drive.params,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json=ANY,
        )


def test_send_file_upload_error(mock_service_with_cookies: PyiCloudService) -> None:
    """Test sending a file to iCloud Drive with an upload error."""
    drive: DriveService = mock_service_with_cookies.drive
    mock_file = Mock()
    mock_file.name = "test_file.txt"
    mock_file.tell = Mock(side_effect=[0, 100, 0])  # Mock file size as 100 bytes

    mock_upload_url_response = [
        {"document_id": "mock_document_id", "url": "https://example.com/upload"}
    ]

    with (
        patch.object(
            drive.session,
            "post",
            side_effect=[
                Mock(
                    ok=True, json=lambda: mock_upload_url_response
                ),  # _get_upload_contentws_url
                Mock(ok=False, reason="Upload Failed"),  # Upload file
            ],
        ) as mock_post,
        patch("mimetypes.guess_type", return_value=("text/plain", None)),
    ):
        with pytest.raises(PyiCloudAPIResponseException, match="Upload Failed"):
            drive.send_file("mock_folder_id", mock_file)

        # Assert _get_upload_contentws_url call
        mock_post.assert_any_call(
            drive._document_root + f"/ws/{CLOUD_DOCS_ZONE}/upload/web",
            params=ANY,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json={
                "filename": "test_file.txt",
                "type": "FILE",
                "content_type": "text/plain",
                "size": 100,
            },
        )

        # Assert file upload call
        mock_post.assert_any_call(
            "https://example.com/upload",
            files={"test_file.txt": mock_file},
        )


def test_send_file_update_error(mock_service_with_cookies: PyiCloudService) -> None:
    """Test sending a file to iCloud Drive with an update error."""
    drive: DriveService = mock_service_with_cookies.drive
    mock_file = Mock()
    mock_file.name = "test_file.txt"
    mock_file.tell = Mock(side_effect=[0, 100, 0])  # Mock file size as 100 bytes

    mock_upload_url_response = [
        {"document_id": "mock_document_id", "url": "https://example.com/upload"}
    ]
    mock_upload_response = {
        "singleFile": {
            "fileChecksum": "mock_checksum",
            "wrappingKey": "mock_key",
            "referenceChecksum": "mock_reference",
            "size": 100,
        }
    }

    with (
        patch.object(
            drive.session,
            "post",
            side_effect=[
                Mock(
                    ok=True, json=lambda: mock_upload_url_response
                ),  # _get_upload_contentws_url
                Mock(ok=True, json=lambda: mock_upload_response),  # Upload file
                Mock(ok=False, reason="Update Failed"),  # _update_contentws
            ],
        ) as mock_post,
        patch("mimetypes.guess_type", return_value=("text/plain", None)),
    ):
        with pytest.raises(PyiCloudAPIResponseException, match="Update Failed"):
            drive.send_file("mock_folder_id", mock_file)

        # Assert _get_upload_contentws_url call
        mock_post.assert_any_call(
            drive._document_root + f"/ws/{CLOUD_DOCS_ZONE}/upload/web",
            params=ANY,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json={
                "filename": "test_file.txt",
                "type": "FILE",
                "content_type": "text/plain",
                "size": 100,
            },
        )

        # Assert file upload call
        mock_post.assert_any_call(
            "https://example.com/upload",
            files={"test_file.txt": mock_file},
        )

        # Assert _update_contentws call
        mock_post.assert_any_call(
            drive._document_root + f"/ws/{CLOUD_DOCS_ZONE}/update/documents",
            params=drive.params,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json=ANY,
        )



"""Test constants."""

from .const_account_family import APPLE_ID_EMAIL, ICLOUD_ID_EMAIL, PRIMARY_EMAIL

# Base
AUTHENTICATED_USER = PRIMARY_EMAIL
REQUIRES_2FA_TOKEN = "requires_2fa_token"
REQUIRES_2FA_USER = "requires_2fa_user"
VALID_USERS = [AUTHENTICATED_USER, REQUIRES_2FA_USER, APPLE_ID_EMAIL, ICLOUD_ID_EMAIL]
VALID_PASSWORD = "valid_password"
VALID_COOKIE = "valid_cookie"
VALID_TOKEN = "valid_token"
VALID_2FA_CODE = "000000"
VALID_TOKENS = [VALID_TOKEN, REQUIRES_2FA_TOKEN]

CLIENT_ID = "client_id"



"""Pyicloud Session handling"""

import logging
import os
from json import JSONDecodeError, dump, load
from os import path
from re import match
from typing import TYPE_CHECKING, Any, NoReturn, Optional, Union, cast

import requests
from requests.models import Response

from pyicloud.const import (
    CONTENT_TYPE,
    CONTENT_TYPE_JSON,
    CONTENT_TYPE_TEXT_JSON,
    ERROR_ACCESS_DENIED,
    ERROR_AUTHENTICATION_FAILED,
    ERROR_ZONE_NOT_FOUND,
    HEADER_DATA,
    AppleAuthError,
)
from pyicloud.cookie_jar import PyiCloudCookieJar
from pyicloud.exceptions import (
    PyiCloud2FARequiredException,
    PyiCloud2SARequiredException,
    PyiCloudAPIResponseException,
    PyiCloudAuthRequiredException,
    PyiCloudServiceNotActivatedException,
)

if TYPE_CHECKING:
    from pyicloud.base import PyiCloudService


class PyiCloudSession(requests.Session):
    """iCloud session."""

    def __init__(
        self,
        service: "PyiCloudService",
        client_id: str,
        cookie_directory: str,
        verify: bool = False,
        headers: Optional[dict[str, str]] = None,
    ) -> None:
        super().__init__()

        self._service: PyiCloudService = service
        self.verify = verify
        self._cookie_directory: str = cookie_directory
        self.cookies = PyiCloudCookieJar(filename=self.cookiejar_path)
        self._data: dict[str, Any] = {}

        self._logger: logging.Logger = logging.getLogger(__name__)

        if headers:
            self.headers.update(headers)

        self._load_session_data()

        if not self._data.get("client_id"):
            self._data.update({"client_id": client_id})

    @property
    def data(self) -> dict[str, Any]:
        """Gets the session data"""
        return self._data

    @property
    def logger(self) -> logging.Logger:
        """Gets the request logger"""
        return self._logger

    def _load_session_data(self) -> None:
        """Load session_data from file."""
        if os.path.exists(self.cookiejar_path):
            try:
                cast(PyiCloudCookieJar, self.cookies).load()
            except (OSError, ValueError) as exc:
                self._logger.warning(
                    "Failed to load cookie jar %s: %s; starting without persisted cookies",
                    self.cookiejar_path,
                    exc,
                )
                cast(PyiCloudCookieJar, self.cookies).clear()

        self._logger.debug("Using session file %s", self.session_path)
        self._data: dict[str, Any] = {}
        try:
            with open(self.session_path, encoding="utf-8") as session_f:
                self._data = load(session_f)
        except (
            JSONDecodeError,
            OSError,
        ):
            self._logger.info("Session file does not exist")

    def _save_session_data(self) -> None:
        """Save session_data to file."""
        if self._cookie_directory and not os.path.isdir(self._cookie_directory):
            os.makedirs(self._cookie_directory, exist_ok=True)
        with open(self.session_path, "w", encoding="utf-8") as outfile:
            dump(self._data, outfile)
            self.logger.debug("Saved session data to file: %s", self.session_path)

        try:
            cast(PyiCloudCookieJar, self.cookies).save()
            self.logger.debug("Saved cookies data to file: %s", self.cookiejar_path)
        except (OSError, ValueError) as exc:
            self.logger.warning("Failed to save cookies data: %s", exc)

    def _update_session_data(self, response: Response) -> None:
        """Update session_data with new data."""
        for header, value in HEADER_DATA.items():
            if response.headers.get(header):
                session_arg: str = value
                self._data.update({session_arg: response.headers.get(header)})

    def _is_json_response(self, response: Response) -> bool:
        content_type: str = response.headers.get(CONTENT_TYPE, "")
        json_mimetypes: list[str] = [
            CONTENT_TYPE_JSON,
            CONTENT_TYPE_TEXT_JSON,
        ]
        return content_type.split(";")[0] in json_mimetypes

    def request(
        self,
        method,
        url,
        params=None,
        data=None,
        headers=None,
        cookies=None,
        files=None,
        auth=None,
        timeout=None,
        allow_redirects=True,
        proxies=None,
        hooks=None,
        stream=None,
        verify=None,
        cert=None,
        json=None,
    ) -> Response:
        return self._request(
            method,
            url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
        )

    def _request(
        self,
        method,
        url,
        **kwargs,
    ) -> Response:
        """Request method."""
        self.logger.debug(
            "%s %s",
            method,
            url,
        )

        try:
            response: Response = super().request(
                method=method,
                url=url,
                **kwargs,
            )

            self._update_session_data(response)
            self._save_session_data()

            status_code: int = int(response.status_code)

            if not response.ok and (
                self._is_json_response(response)
                or status_code
                in [
                    AppleAuthError.TWO_FACTOR_REQUIRED,
                    AppleAuthError.FIND_MY_REAUTH_REQUIRED,
                    AppleAuthError.LOGIN_TOKEN_EXPIRED,
                    AppleAuthError.GENERAL_AUTH_ERROR,
                ],
            ):
                return self._handle_request_error(
                    status_code=status_code,
                    response=response,
                )

            response.raise_for_status()

            if not self._is_json_response(response):
                return response

            self._decode_json_response(response)

            return response
        except requests.HTTPError as err:
            raise PyiCloudAPIResponseException(
                reason=err.response.text,
                code=err.response.status_code,
            ) from err
        except requests.exceptions.RequestException as err:
            raise PyiCloudAPIResponseException("Request failed to iCloud") from err

    def _handle_request_error(
        self,
        status_code: int,
        response: Response,
    ) -> Response:
        """Handle request error."""
        if (
            status_code == AppleAuthError.TWO_FACTOR_REQUIRED
            and self._is_json_response(response)
            and (response.json().get("authType") == "hsa2")
        ):
            raise PyiCloud2FARequiredException(
                apple_id=self.service.account_name,
                response=response,
            )

        if status_code == AppleAuthError.FIND_MY_REAUTH_REQUIRED:
            raise PyiCloudAuthRequiredException(
                apple_id=self.service.account_name,
                response=response,
            )

        self._raise_error(response, status_code, response.reason)

    def _decode_json_response(self, response: Response) -> None:
        """Decode JSON response."""
        if len(response.content) == 0:
            return

        try:
            data: Union[list[dict[str, Any]], dict[str, Any]] = response.json()
            if isinstance(data, dict):
                reason: Optional[str] = data.get("errorMessage")
                reason = reason or data.get("reason")
                reason = reason or data.get("errorReason")
                reason = reason or data.get("error")
                if reason and not isinstance(reason, str):
                    reason = "Unknown reason"

                if reason:
                    code: Optional[Union[int, str]] = data.get("errorCode")
                    code = code or data.get("serverErrorCode")
                    self._raise_error(response, code, reason)

        except JSONDecodeError:
            self.logger.warning(
                "Failed to parse response with JSON mimetype: %s", response.text
            )

    def _raise_error(
        self, response: Response, code: Optional[Union[int, str]], reason: str
    ) -> NoReturn:
        if (
            self.service.requires_2sa
            and reason == "Missing X-APPLE-WEBAUTH-TOKEN cookie"
        ):
            raise PyiCloud2SARequiredException(self.service.account_name)
        if code in (ERROR_ZONE_NOT_FOUND, ERROR_AUTHENTICATION_FAILED):
            reason = (
                "Please log into https://icloud.com/ to manually "
                "finish setting up your iCloud service"
            )
            raise PyiCloudServiceNotActivatedException(reason, code, response)
        if code == ERROR_ACCESS_DENIED:
            reason = (
                reason + ".  Please wait a few minutes then try again."
                "The remote servers might be trying to throttle requests."
            )
        if isinstance(code, int) and code in [
            AppleAuthError.TWO_FACTOR_REQUIRED,
            AppleAuthError.FIND_MY_REAUTH_REQUIRED,
            AppleAuthError.LOGIN_TOKEN_EXPIRED,
            AppleAuthError.GENERAL_AUTH_ERROR,
        ]:
            reason = "Authentication required for Account."

        raise PyiCloudAPIResponseException(reason, code, response)

    @property
    def service(self) -> "PyiCloudService":
        """Gets the service."""
        return self._service

    @property
    def cookiejar_path(self) -> str:
        """Get path for cookiejar file."""
        return path.join(
            self._cookie_directory,
            "".join([c for c in self.service.account_name if match(r"\w", c)])
            + ".cookiejar",
        )

    @property
    def session_path(self) -> str:
        """Get path for session data file."""
        return path.join(
            self._cookie_directory,
            "".join([c for c in self.service.account_name if match(r"\w", c)])
            + ".session",
        )



"""Tests for the Hide My Email service."""
# pylint: disable=protected-access

from typing import Any, Optional
from unittest.mock import MagicMock

from requests import Response

from pyicloud.services.hidemyemail import HideMyEmailService


def test_generate(
    hidemyemail_service: HideMyEmailService, mock_session: MagicMock
) -> None:
    """Test the generate method."""
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"result": {"hme": "alias@example.com"}}
    mock_session.post.return_value = mock_response

    result: Optional[str] = hidemyemail_service.generate()
    assert result == "alias@example.com"
    mock_session.post.assert_called_once_with(
        "https://example.com/v1/hme/generate", params={"dsid": "12345"}
    )


def test_reserve(
    hidemyemail_service: HideMyEmailService, mock_session: MagicMock
) -> None:
    """Test the reserve method."""
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"result": {"status": "success"}}
    mock_session.post.return_value = mock_response

    result: dict[str, Any] = hidemyemail_service.reserve(
        "alias@example.com", "Test Label", "Test Note"
    )
    assert result == {"status": "success"}
    mock_session.post.assert_called_once_with(
        "https://example.com/v1/hme/reserve",
        params={"dsid": "12345"},
        json={"hme": "alias@example.com", "label": "Test Label", "note": "Test Note"},
    )


def test_len(hidemyemail_service: HideMyEmailService, mock_session: MagicMock) -> None:
    """Test the __len__ method."""
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"result": {"hmeEmails": ["email1", "email2"]}}
    mock_session.get.return_value = mock_response

    result: int = len(hidemyemail_service)
    assert result == 2
    mock_session.get.assert_called_once_with(
        "https://example.com/v2/hme/list", params={"dsid": "12345"}
    )


def test_iter(hidemyemail_service: HideMyEmailService, mock_session: MagicMock) -> None:
    """Test the __iter__ method."""
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"result": {"hmeEmails": ["email1", "email2"]}}
    mock_session.get.return_value = mock_response

    emails = list(iter(hidemyemail_service))
    assert emails == ["email1", "email2"]
    mock_session.get.assert_called_once_with(
        "https://example.com/v2/hme/list", params={"dsid": "12345"}
    )


def test_getitem(
    hidemyemail_service: HideMyEmailService, mock_session: MagicMock
) -> None:
    """Test the __getitem__ method."""
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"result": {"email": "alias@example.com"}}
    mock_session.post.return_value = mock_response

    result: dict[str, Any] = hidemyemail_service["12345"]
    assert result == {"email": "alias@example.com"}
    mock_session.post.assert_called_once_with(
        "https://example.com/v2/hme/get",
        params={"dsid": "12345"},
        json={"anonymousId": "12345"},
    )


def test_update_metadata(
    hidemyemail_service: HideMyEmailService, mock_session: MagicMock
) -> None:
    """Test the update_metadata method."""
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"result": {"status": "updated"}}
    mock_session.post.return_value = mock_response

    result: dict[str, Any] = hidemyemail_service.update_metadata(
        "12345", "New Label", "New Note"
    )
    assert result == {"status": "updated"}
    mock_session.post.assert_called_once_with(
        "https://example.com/v1/hme/updateMetaData",
        params={"dsid": "12345"},
        json={"anonymousId": "12345", "label": "New Label", "note": "New Note"},
    )


def test_delete(
    hidemyemail_service: HideMyEmailService, mock_session: MagicMock
) -> None:
    """Test the delete method."""
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"result": {"status": "deleted"}}
    mock_session.post.return_value = mock_response

    result: dict[str, Any] = hidemyemail_service.delete("12345")
    assert result == {"status": "deleted"}
    mock_session.post.assert_called_once_with(
        "https://example.com/v1/hme/delete",
        params={"dsid": "12345"},
        json={"anonymousId": "12345"},
    )


def test_deactivate(
    hidemyemail_service: HideMyEmailService, mock_session: MagicMock
) -> None:
    """Test the deactivate method."""
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"result": {"status": "deactivated"}}
    mock_session.post.return_value = mock_response

    result: dict[str, Any] = hidemyemail_service.deactivate("12345")
    assert result == {"status": "deactivated"}
    mock_session.post.assert_called_once_with(
        "https://example.com/v1/hme/deactivate",
        params={"dsid": "12345"},
        json={"anonymousId": "12345"},
    )


def test_reactivate(
    hidemyemail_service: HideMyEmailService, mock_session: MagicMock
) -> None:
    """Test the reactivate method."""
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"result": {"status": "reactivated"}}
    mock_session.post.return_value = mock_response

    result: dict[str, Any] = hidemyemail_service.reactivate("12345")
    assert result == {"status": "reactivated"}
    mock_session.post.assert_called_once_with(
        "https://example.com/v1/hme/reactivate",
        params={"dsid": "12345"},
        json={"anonymousId": "12345"},
    )



"""Contacts service."""

from typing import Any, Optional

from requests import Response

from pyicloud.services.base import BaseService
from pyicloud.session import PyiCloudSession


class ContactsService(BaseService):
    """
    The 'Contacts' iCloud service, connects to iCloud and returns contacts.
    """

    def __init__(
        self, service_root: str, session: PyiCloudSession, params: dict[str, Any]
    ) -> None:
        super().__init__(service_root, session, params)
        self._contacts_endpoint: str = f"{self.service_root}/co"
        self._contacts_refresh_url: str = f"{self._contacts_endpoint}/startup"
        self._contacts_next_url: str = f"{self._contacts_endpoint}/contacts"
        self._contacts_changeset_url: str = f"{self._contacts_endpoint}/changeset"
        self._contacts_me_card_url: str = f"{self._contacts_endpoint}/mecard"

        self._contacts: Optional[list[dict[str, Any]]] = None

    def refresh_client(self) -> None:
        """
        Refreshes the ContactsService endpoint, ensuring that the
        contacts data is up-to-date.
        """
        params_contacts: dict[str, Any] = dict(self.params)
        params_contacts.update(
            {
                "locale": "en_US",
                "order": "last,first",
                "includePhoneNumbers": True,
                "includePhotos": True,
            }
        )
        req: Response = self.session.get(
            self._contacts_refresh_url, params=params_contacts
        )
        response: dict[str, Any] = req.json()

        params_next: dict[str, Any] = dict(params_contacts)
        params_next.update(
            {
                "prefToken": response["prefToken"],
                "syncToken": response["syncToken"],
                "limit": "0",
                "offset": "0",
            }
        )
        req = self.session.get(self._contacts_next_url, params=params_next)
        response = req.json()
        self._contacts = response.get("contacts")

    @property
    def all(self) -> list[dict[str, Any]] | None:
        """
        Retrieves all contacts.
        """
        self.refresh_client()
        return self._contacts

    @property
    def me(self) -> "MeCard":
        """
        Retrieves the user's own contact information.
        """
        params_contacts = dict(self.params)
        req: Response = self.session.get(
            self._contacts_me_card_url, params=params_contacts
        )
        response = req.json()
        return MeCard(response)


class MeCard:
    """
    The 'MeCard' class represents the user's own contact information.
    """

    def __init__(self, data: dict[str, Any]) -> None:
        self._data: dict[str, Any] = data
        contacts = data.get("contacts")
        if isinstance(contacts, list) and isinstance(contacts[0], dict):
            self._contact: dict[str, Any] = contacts[0]
        else:
            raise KeyError("contacts not found in data")

    @property
    def first_name(self) -> str:
        """
        The user's first name.
        """
        return self._contact["firstName"]

    @property
    def last_name(self) -> str:
        """
        The user's last name.
        """
        return self._contact["lastName"]

    @property
    def photo(self):
        """
        The user's photo.
        """
        return self._contact["photo"]

    def __str__(self) -> str:
        return f"{self.first_name} {self.last_name}"

    def __repr__(self) -> str:
        return f"<MeCard({self.first_name}-{self.last_name})>"

    @property
    def raw_data(self) -> dict[str, Any]:
        """
        The raw data of the mecard.
        """
        return self._data



"""Drive test constants."""
# pylint: disable=line-too-long

DRIVEWSID = "FOLDER::com.apple.CloudDocs::root"
ZONE = "com.apple.CloudDocs"
DATE_CREATED = "2019-12-12T14:33:55-08:00"
FOLDER1 = "FOLDER::com.apple.CloudDocs::1C7F1760-D940-480F-8C4F-005824A4E05B"
FOLDER2 = "FOLDER::com.apple.CloudDocs::D5AA0425-E84F-4501-AF5D-60F1D92648CF"
# Data
DRIVE_ROOT_WORKING = [
    {
        "drivewsid": DRIVEWSID,
        "docwsid": "root",
        "zone": ZONE,
        "name": "",
        "etag": "31",
        "type": "FOLDER",
        "assetQuota": 62418076,
        "fileCount": 7,
        "shareCount": 0,
        "shareAliasCount": 0,
        "directChildrenCount": 3,
        "items": [
            {
                "dateCreated": DATE_CREATED,
                "drivewsid": "FOLDER::com.apple.Keynote::documents",
                "docwsid": "documents",
                "zone": "com.apple.Keynote",
                "name": "Keynote",
                "parentId": DRIVEWSID,
                "etag": "2m",
                "type": "APP_LIBRARY",
                "maxDepth": "ANY",
                "icons": [
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Keynote&field=icon120x120_iOS",
                        "type": "IOS",
                        "size": 120,
                    },
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Keynote&field=icon80x80_iOS",
                        "type": "IOS",
                        "size": 80,
                    },
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Keynote&field=icon40x40_iOS",
                        "type": "IOS",
                        "size": 40,
                    },
                ],
                "supportedExtensions": [
                    "pptx",
                    "ppsx",
                    "pps",
                    "pot",
                    "key-tef",
                    "ppt",
                    "potx",
                    "potm",
                    "pptm",
                    "ppsm",
                    "key",
                    "kth",
                ],
                "supportedTypes": [
                    "com.microsoft.powerpoint.pps",
                    "com.microsoft.powerpoint.pot",
                    "com.microsoft.powerpoint.ppt",
                    "org.openxmlformats.presentationml.template.macroenabled",
                    "org.openxmlformats.presentationml.slideshow.macroenabled",
                    "com.apple.iwork.keynote.key-tef",
                    "org.openxmlformats.presentationml.template",
                    "org.openxmlformats.presentationml.presentation.macroenabled",
                    "com.apple.iwork.keynote.key",
                    "com.apple.iwork.keynote.kth",
                    "org.openxmlformats.presentationml.presentation",
                    "org.openxmlformats.presentationml.slideshow",
                    "com.apple.iwork.keynote.sffkey",
                    "com.apple.iwork.keynote.sffkth",
                ],
            },
            {
                "dateCreated": DATE_CREATED,
                "drivewsid": "FOLDER::com.apple.Numbers::documents",
                "docwsid": "documents",
                "zone": "com.apple.Numbers",
                "name": "Numbers",
                "parentId": DRIVEWSID,
                "etag": "3k",
                "type": "APP_LIBRARY",
                "maxDepth": "ANY",
                "icons": [
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Numbers&field=icon120x120_iOS",
                        "type": "IOS",
                        "size": 120,
                    },
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Numbers&field=icon80x80_iOS",
                        "type": "IOS",
                        "size": 80,
                    },
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Numbers&field=icon40x40_iOS",
                        "type": "IOS",
                        "size": 40,
                    },
                ],
                "supportedExtensions": [
                    "hh",
                    "ksh",
                    "lm",
                    "xlt",
                    "c++",
                    "f95",
                    "lid",
                    "csv",
                    "numbers",
                    "php4",
                    "hp",
                    "py",
                    "nmbtemplate",
                    "lmm",
                    "jscript",
                    "php3",
                    "crash",
                    "patch",
                    "java",
                    "ym",
                    "xlam",
                    "text",
                    "mi",
                    "exp",
                    "adb",
                    "jav",
                    "ada",
                    "ii",
                    "defs",
                    "mm",
                    "cpp",
                    "cxx",
                    "pas",
                    "diff",
                    "pch++",
                    "javascript",
                    "panic",
                    "rb",
                    "ads",
                    "tcsh",
                    "ypp",
                    "yxx",
                    "ph3",
                    "ph4",
                    "phtml",
                    "xltx",
                    "hang",
                    "rbw",
                    "f77",
                    "for",
                    "js",
                    "h++",
                    "mig",
                    "gpurestart",
                    "mii",
                    "zsh",
                    "m3u",
                    "pch",
                    "sh",
                    "xltm",
                    "applescript",
                    "tsv",
                    "ymm",
                    "shutdownstall",
                    "cc",
                    "xlsx",
                    "scpt",
                    "c",
                    "inl",
                    "f",
                    "numbers-tef",
                    "h",
                    "i",
                    "hpp",
                    "hxx",
                    "dlyan",
                    "xla",
                    "l",
                    "cp",
                    "m",
                    "lpp",
                    "lxx",
                    "txt",
                    "r",
                    "s",
                    "xlsm",
                    "spin",
                    "php",
                    "csh",
                    "y",
                    "bash",
                    "m3u8",
                    "pl",
                    "f90",
                    "pm",
                    "xls",
                ],
                "supportedTypes": [
                    "org.openxmlformats.spreadsheetml.sheet",
                    "com.microsoft.excel.xla",
                    "com.apple.iwork.numbers.template",
                    "org.openxmlformats.spreadsheetml.sheet.macroenabled",
                    "com.apple.iwork.numbers.sffnumbers",
                    "com.apple.iwork.numbers.numbers",
                    "public.plain-text",
                    "com.microsoft.excel.xlt",
                    "org.openxmlformats.spreadsheetml.template",
                    "com.microsoft.excel.xls",
                    "public.comma-separated-values-text",
                    "com.apple.iwork.numbers.numbers-tef",
                    "org.openxmlformats.spreadsheetml.template.macroenabled",
                    "public.tab-separated-values-text",
                    "com.apple.iwork.numbers.sfftemplate",
                    "com.microsoft.excel.openxml.addin",
                ],
            },
            {
                "dateCreated": DATE_CREATED,
                "drivewsid": "FOLDER::com.apple.Pages::documents",
                "docwsid": "documents",
                "zone": "com.apple.Pages",
                "name": "Pages",
                "parentId": DRIVEWSID,
                "etag": "km",
                "type": "APP_LIBRARY",
                "maxDepth": "ANY",
                "icons": [
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Pages&field=icon120x120_iOS",
                        "type": "IOS",
                        "size": 120,
                    },
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Pages&field=icon80x80_iOS",
                        "type": "IOS",
                        "size": 80,
                    },
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Pages&field=icon40x40_iOS",
                        "type": "IOS",
                        "size": 40,
                    },
                ],
                "supportedExtensions": [
                    "hh",
                    "ksh",
                    "lm",
                    "c++",
                    "f95",
                    "lid",
                    "php4",
                    "hp",
                    "py",
                    "lmm",
                    "jscript",
                    "php3",
                    "crash",
                    "patch",
                    "pages",
                    "java",
                    "ym",
                    "text",
                    "mi",
                    "exp",
                    "adb",
                    "jav",
                    "ada",
                    "ii",
                    "defs",
                    "mm",
                    "cpp",
                    "cxx",
                    "pas",
                    "pages-tef",
                    "diff",
                    "pch++",
                    "javascript",
                    "panic",
                    "rb",
                    "ads",
                    "tcsh",
                    "rtfd",
                    "ypp",
                    "yxx",
                    "doc",
                    "ph3",
                    "ph4",
                    "template",
                    "phtml",
                    "hang",
                    "rbw",
                    "f77",
                    "dot",
                    "for",
                    "js",
                    "h++",
                    "mig",
                    "gpurestart",
                    "mii",
                    "zsh",
                    "m3u",
                    "pch",
                    "sh",
                    "applescript",
                    "ymm",
                    "shutdownstall",
                    "dotx",
                    "cc",
                    "scpt",
                    "c",
                    "rtf",
                    "inl",
                    "f",
                    "h",
                    "i",
                    "hpp",
                    "hxx",
                    "dlyan",
                    "l",
                    "cp",
                    "m",
                    "lpp",
                    "lxx",
                    "docx",
                    "txt",
                    "r",
                    "s",
                    "spin",
                    "php",
                    "csh",
                    "y",
                    "bash",
                    "m3u8",
                    "pl",
                    "f90",
                    "pm",
                ],
                "supportedTypes": [
                    "com.apple.rtfd",
                    "com.apple.iwork.pages.sffpages",
                    "com.apple.iwork.pages.sfftemplate",
                    "com.microsoft.word.dot",
                    "com.apple.iwork.pages.pages",
                    "com.microsoft.word.doc",
                    "org.openxmlformats.wordprocessingml.template",
                    "org.openxmlformats.wordprocessingml.document",
                    "com.apple.iwork.pages.pages-tef",
                    "com.apple.iwork.pages.template",
                    "public.rtf",
                    "public.plain-text",
                ],
            },
            {
                "dateCreated": DATE_CREATED,
                "drivewsid": "FOLDER::com.apple.Preview::documents",
                "docwsid": "documents",
                "zone": "com.apple.Preview",
                "name": "Preview",
                "parentId": DRIVEWSID,
                "etag": "bv",
                "type": "APP_LIBRARY",
                "maxDepth": "ANY",
                "icons": [
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Preview&field=icon32x32_OSX",
                        "type": "OSX",
                        "size": 32,
                    },
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Preview&field=icon128x128_OSX",
                        "type": "OSX",
                        "size": 128,
                    },
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Preview&field=icon16x16_OSX",
                        "type": "OSX",
                        "size": 16,
                    },
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Preview&field=icon256x256_OSX",
                        "type": "OSX",
                        "size": 256,
                    },
                    {
                        "url": "https://p31-drivews.icloud.com/getIcons?id=com.apple.Preview&field=icon64x64_OSX",
                        "type": "OSX",
                        "size": 64,
                    },
                ],
                "supportedExtensions": [
                    "ps",
                    "nmbtemplate",
                    "astc",
                    "mpkg",
                    "prefpane",
                    "pef",
                    "mos",
                    "qlgenerator",
                    "scptd",
                    "raf",
                    "saver",
                    "band",
                    "dng",
                    "pict",
                    "exr",
                    "kth",
                    "appex",
                    "app",
                    "pages-tef",
                    "slidesaver",
                    "pluginkit",
                    "distz",
                    "ai",
                    "png",
                    "eps",
                    "raw",
                    "pvr",
                    "mpo",
                    "ktx",
                    "nrw",
                    "lpdf",
                    "pfm",
                    "3fr",
                    "template",
                    "imovielibrary",
                    "pwl",
                    "iwwebpackage",
                    "wdgt",
                    "tga",
                    "pgm",
                    "erf",
                    "jpeg",
                    "j2c",
                    "bundle",
                    "key",
                    "j2k",
                    "abc",
                    "arw",
                    "xpc",
                    "pic",
                    "ppm",
                    "menu",
                    "icns",
                    "mrw",
                    "plugin",
                    "mdimporter",
                    "bmp",
                    "numbers",
                    "dae",
                    "dist",
                    "pic",
                    "rw2",
                    "nef",
                    "tif",
                    "pages",
                    "sgi",
                    "ico",
                    "theater",
                    "gbproj",
                    "webplugin",
                    "cr2",
                    "fff",
                    "webp",
                    "jp2",
                    "sr2",
                    "rtfd",
                    "pbm",
                    "pkpass",
                    "jfx",
                    "fpbf",
                    "psd",
                    "xbm",
                    "tiff",
                    "avchd",
                    "gif",
                    "pntg",
                    "rwl",
                    "pset",
                    "pkg",
                    "dcr",
                    "hdr",
                    "jpe",
                    "pct",
                    "jpg",
                    "jpf",
                    "orf",
                    "srf",
                    "numbers-tef",
                    "iconset",
                    "crw",
                    "fpx",
                    "dds",
                    "pdf",
                    "jpx",
                    "key-tef",
                    "efx",
                    "hdr",
                    "srw",
                ],
                "supportedTypes": [
                    "com.adobe.illustrator.ai-image",
                    "com.kodak.flashpix-image",
                    "public.pbm",
                    "com.apple.pict",
                    "com.ilm.openexr-image",
                    "com.sgi.sgi-image",
                    "com.apple.icns",
                    "public.heifs",
                    "com.truevision.tga-image",
                    "com.adobe.postscript",
                    "public.camera-raw-image",
                    "public.pvr",
                    "public.png",
                    "com.adobe.photoshop-image",
                    "public.heif",
                    "com.microsoft.ico",
                    "com.adobe.pdf",
                    "public.heic",
                    "public.xbitmap-image",
                    "com.apple.localized-pdf-bundle",
                    "public.3d-content",
                    "com.compuserve.gif",
                    "public.avci",
                    "public.jpeg",
                    "com.apple.rjpeg",
                    "com.adobe.encapsulated-postscript",
                    "com.microsoft.bmp",
                    "public.fax",
                    "org.khronos.astc",
                    "com.apple.application-bundle",
                    "public.avcs",
                    "public.webp",
                    "public.heics",
                    "com.apple.macpaint-image",
                    "public.mpo-image",
                    "public.jpeg-2000",
                    "public.tiff",
                    "com.microsoft.dds",
                    "com.apple.pdf-printer-settings",
                    "org.khronos.ktx",
                    "public.radiance",
                    "com.apple.package",
                    "public.folder",
                ],
            },
            {
                "drivewsid": FOLDER1,
                "docwsid": "1C7F1760-D940-480F-8C4F-005824A4E05B",
                "zone": ZONE,
                "name": "pyiCloud",
                "parentId": DRIVEWSID,
                "etag": "30",
                "type": "FOLDER",
                "assetQuota": 42199575,
                "fileCount": 2,
                "shareCount": 0,
                "shareAliasCount": 0,
                "directChildrenCount": 1,
            },
        ],
        "numberOfItems": 5,
    }
]

# App specific folder (Keynote, Numbers, Pages, Preview ...) type=APP_LIBRARY
DRIVE_ROOT_INVALID = [
    {"drivewsid": "FOLDER::com.apple.CloudDocs::documents", "status": "ID_INVALID"}
]

DRIVE_FOLDER_WORKING = [
    {
        "drivewsid": FOLDER1,
        "docwsid": "1C7F1760-D940-480F-8C4F-005824A4E05B",
        "zone": ZONE,
        "name": "pyiCloud",
        "parentId": DRIVEWSID,
        "etag": "30",
        "type": "FOLDER",
        "assetQuota": 42199575,
        "fileCount": 2,
        "shareCount": 0,
        "shareAliasCount": 0,
        "directChildrenCount": 1,
        "items": [
            {
                "drivewsid": FOLDER2,
                "docwsid": "D5AA0425-E84F-4501-AF5D-60F1D92648CF",
                "zone": ZONE,
                "name": "Test",
                "parentId": FOLDER1,
                "etag": "2z",
                "type": "FOLDER",
                "assetQuota": 42199575,
                "fileCount": 2,
                "shareCount": 0,
                "shareAliasCount": 0,
                "directChildrenCount": 2,
            }
        ],
        "numberOfItems": 1,
    }
]

DRIVE_SUBFOLDER_WORKING = [
    {
        "drivewsid": FOLDER2,
        "docwsid": "D5AA0425-E84F-4501-AF5D-60F1D92648CF",
        "zone": ZONE,
        "name": "Test",
        "parentId": FOLDER1,
        "etag": "2z",
        "type": "FOLDER",
        "assetQuota": 42199575,
        "fileCount": 2,
        "shareCount": 0,
        "shareAliasCount": 0,
        "directChildrenCount": 2,
        "items": [
            {
                "drivewsid": "FILE::com.apple.CloudDocs::33A41112-4131-4938-9691-7F356CE3C51D",
                "docwsid": "33A41112-4131-4938-9691-7F356CE3C51D",
                "zone": ZONE,
                "name": "Document scanné 2",
                "parentId": FOLDER2,
                "dateModified": "2020-04-27T21:37:36Z",
                "dateChanged": "2020-04-27T14:44:29-07:00",
                "size": 19876991,
                "etag": "2k::2j",
                "extension": "pdf",
                "hiddenExtension": True,
                "lastOpenTime": "2020-04-27T21:37:36Z",
                "type": "FILE",
            },
            {
                "drivewsid": "FILE::com.apple.CloudDocs::516C896C-6AA5-4A30-B30E-5502C2333DAE",
                "docwsid": "516C896C-6AA5-4A30-B30E-5502C2333DAE",
                "zone": ZONE,
                "name": "Scanned document 1",
                "parentId": FOLDER2,
                "dateModified": "2020-05-03T00:15:17Z",
                "dateChanged": "2020-05-02T17:16:17-07:00",
                "size": 21644358,
                "etag": "32::2x",
                "extension": "pdf",
                "hiddenExtension": True,
                "lastOpenTime": "2020-05-03T00:24:25Z",
                "type": "FILE",
            },
        ],
        "numberOfItems": 2,
    }
]

DRIVE_FILE_DOWNLOAD_WORKING = {
    "document_id": "516C896C-6AA5-4A30-B30E-5502C2333DAE",
    "data_token": {
        "url": "https://cvws.icloud-content.com/B/signature1ref_signature1/Scanned+document+1.pdf?o=object1&v=1&x=3&a=token1&e=1588472097&k=wrapping_key1&fl=&r=request&ckc=com.apple.clouddocs&ckz=com.apple.CloudDocs&p=31&s=s1",
        "token": "token1",
        "signature": "signature1",
        "wrapping_key": "wrapping_key1==",
        "reference_signature": "ref_signature1",
    },
    "thumbnail_token": {
        "url": "https://cvws.icloud-content.com/B/signature2ref_signature2/Scanned+document+1.jpg?o=object2&v=1&x=3&a=token2&e=1588472097&k=wrapping_key2&fl=&r=request&ckc=com.apple.clouddocs&ckz=com.apple.CloudDocs&p=31&s=s2",
        "token": "token2",
        "signature": "signature2",
        "wrapping_key": "wrapping_key2==",
        "reference_signature": "ref_signature2",
    },
    "double_etag": "32::2x",
}

DRIVE_TRASH_WORKING = [
    {
        "items": [
            {
                "dateCreated": "2022-06-23T20:58:35Z",
                "drivewsid": "FILE::com.apple.CloudDocs::C2AD01E4-E625-47FE-AE83-4DF311A05A48",
                "docwsid": "C2AD01E4-E625-47FE-AE83-4DF311A05A48",
                "zone": ZONE,
                "name": "dead-file",
                "extension": "download",
                "parentId": "TRASH_ROOT",
                "dateExpiration": "2024-12-12T02:17:55Z",
                "isChainedToParent": True,
                "dateModified": "2022-06-23T20:43:02Z",
                "dateChanged": "2024-11-12T02:17:55Z",
                "size": 11364977,
                "etag": "o72::o6y",
                "restorePath": "Downloads/dead-file.download",
                "lastOpenTime": "2024-11-12T02:15:18Z",
                "type": "FILE",
            },
            {
                "dateCreated": "2024-11-12T04:41:18Z",
                "drivewsid": "FOLDER::com.apple.CloudDocs::31102B37-D62F-4322-862C-EDE2030C8AFA",
                "docwsid": "31102B37-D62F-4322-862C-EDE2030C8AFA",
                "zone": ZONE,
                "name": "test_create_folder",
                "parentId": "TRASH_ROOT",
                "dateExpiration": "2024-12-12T04:48:22Z",
                "isChainedToParent": True,
                "restorePath": "test_create_folder",
                "etag": "o96",
                "type": "FOLDER",
                "assetQuota": 0,
                "fileCount": 0,
                "shareCount": 0,
                "shareAliasCount": 0,
                "directChildrenCount": 0,
            },
            {
                "dateCreated": "2024-11-12T04:18:13Z",
                "drivewsid": "FOLDER::com.apple.CloudDocs::478AEA23-42A2-468A-ABC1-1A04BC07F738",
                "docwsid": "478AEA23-42A2-468A-ABC1-1A04BC07F738",
                "zone": ZONE,
                "name": "test_delete_forever_and_ever",
                "parentId": "TRASH_ROOT",
                "dateExpiration": "2024-12-12T04:18:20Z",
                "isChainedToParent": True,
                "restorePath": "test_delete_forever_and_ever",
                "etag": "o8h",
                "type": "FOLDER",
                "assetQuota": 0,
                "fileCount": 0,
                "shareCount": 0,
                "shareAliasCount": 0,
                "directChildrenCount": 0,
            },
            {
                "dateCreated": "2024-11-12T03:41:18Z",
                "drivewsid": "FOLDER::com.apple.CloudDocs::E63A9193-4428-4AE1-A334-83B880C75379",
                "docwsid": "E63A9193-4428-4AE1-A334-83B880C75379",
                "zone": ZONE,
                "name": "test_files_1",
                "parentId": "TRASH_ROOT",
                "dateExpiration": "2024-12-12T03:42:07Z",
                "isChainedToParent": True,
                "restorePath": "test_files_1",
                "etag": "o7s",
                "type": "FOLDER",
                "assetQuota": 7,
                "fileCount": 1,
                "shareCount": 0,
                "shareAliasCount": 0,
                "directChildrenCount": 1,
            },
            {
                "dateCreated": "2024-11-12T03:37:13Z",
                "drivewsid": "FOLDER::com.apple.CloudDocs::2BF8600B-5DCC-4421-805A-1C28D07197D5",
                "docwsid": "2BF8600B-5DCC-4421-805A-1C28D07197D5",
                "zone": ZONE,
                "name": "test_random_uuid",
                "parentId": "TRASH_ROOT",
                "dateExpiration": "2024-12-12T03:57:30Z",
                "isChainedToParent": True,
                "restorePath": "test_random_uuid",
                "etag": "o9a",
                "type": "FOLDER",
                "assetQuota": 0,
                "fileCount": 0,
                "shareCount": 0,
                "shareAliasCount": 0,
                "directChildrenCount": 0,
            },
            {
                "dateCreated": "2024-11-12T04:25:27Z",
                "drivewsid": "FOLDER::com.apple.CloudDocs::B9B90B8D-CCC2-4BDB-A58D-289F746C3478",
                "docwsid": "B9B90B8D-CCC2-4BDB-A58D-289F746C3478",
                "zone": ZONE,
                "name": "test12345",
                "parentId": "TRASH_ROOT",
                "dateExpiration": "2024-12-12T04:31:46Z",
                "isChainedToParent": True,
                "restorePath": "test12345",
                "etag": "o8y",
                "type": "FOLDER",
                "assetQuota": 0,
                "fileCount": 0,
                "shareCount": 0,
                "shareAliasCount": 0,
                "directChildrenCount": 0,
            },
        ],
        "numberOfItems": 6,
        "drivewsid": "TRASH_ROOT",
    }
]

DRIVE_TRASH_RECOVER_WORKING = {
    "items": [
        {
            "dateCreated": "2024-11-12T03:37:13Z",
            "drivewsid": "FOLDER::com.apple.CloudDocs::2BF8600B-5DCC-4421-805A-1C28D07197D5",
            "docwsid": "2BF8600B-5DCC-4421-805A-1C28D07197D5",
            "zone": ZONE,
            "name": "test_random_uuid",
            "parentId": DRIVEWSID,
            "isChainedToParent": True,
            "item_id": "CJC_vaYFEAAiEH8Y2nkmm0bfntz-AmIQWC4",
            "etag": "o9g",
            "type": "FOLDER",
            "assetQuota": 0,
            "fileCount": 0,
            "shareCount": 0,
            "shareAliasCount": 0,
            "directChildrenCount": 0,
            "status": "OK",
        }
    ]
}

DRIVE_TRASH_DELETE_FOREVER_WORKING = {
    "items": [
        {
            "dateCreated": "2024-11-12T04:18:14Z",
            "drivewsid": "FOLDER::com.apple.CloudDocs::478AEA23-42A2-468A-ABC1-1A04BC07F738",
            "docwsid": "478AEA23-42A2-468A-ABC1-1A04BC07F738",
            "zone": ZONE,
            "name": "test_delete_forever_and_ever",
            "isDeleted": True,
            "parentId": "FOLDER::com.apple.CloudDocs::43D7C666-6E6E-4522-8999-0B519C3A1F4B",
            "dateExpiration": "2024-12-12T04:18:20Z",
            "isChainedToParent": True,
            "item_id": "CJqQty4QACIQjiS90WklSeGExLvHPWWruzgB",
            "restorePath": "test_delete_forever_and_ever",
            "etag": "null",
            "type": "FOLDER",
            "assetQuota": 0,
            "fileCount": 0,
            "shareCount": 0,
            "shareAliasCount": 0,
            "directChildrenCount": 0,
            "status": "OK",
        }
    ]
}



"""SRP password handling."""

from enum import Enum
from hashlib import pbkdf2_hmac, sha256


class SrpProtocolType(Enum):
    """SRP password types."""

    S2K = "s2k"
    S2K_FO = "s2k_fo"


class SrpPassword:
    """SRP password."""

    def __init__(self, password: str) -> None:
        self._password_hash: bytes = sha256(password.encode("utf-8")).digest()
        self.salt: bytes | None = None
        self.iterations: int | None = None
        self.key_length: int | None = None
        self.protocol: SrpProtocolType | None = None

    def set_encrypt_info(
        self, salt: bytes, iterations: int, key_length: int, protocol: SrpProtocolType
    ) -> None:
        """Set encrypt info."""
        self.salt = salt
        self.iterations = iterations
        self.key_length = key_length
        self.protocol = protocol

    def encode(self) -> bytes:
        """Encode password."""
        if self.salt is None or self.iterations is None or self.key_length is None:
            raise ValueError("Encrypt info not set")

        password_digest: bytes | None = None

        if self.protocol == SrpProtocolType.S2K_FO:
            password_digest = self._password_hash.hex().encode()
        elif self.protocol == SrpProtocolType.S2K:
            password_digest = self._password_hash

        if password_digest is None:
            raise ValueError("Unsupported SrpPassword type")

        return pbkdf2_hmac(
            "sha256",
            password_digest,
            self.salt,
            self.iterations,
            self.key_length,
        )



"""Unit tests for the ContactsService and MeCard classes."""
# pylint: disable=protected-access

from unittest.mock import MagicMock, patch

import pytest

from pyicloud.services.contacts import ContactsService, MeCard


def test_contacts_service_initialization(contacts_service: ContactsService) -> None:
    """Test the initialization of ContactsService."""
    assert contacts_service._contacts_endpoint == "https://example.com/co"
    assert contacts_service._contacts_refresh_url == "https://example.com/co/startup"
    assert contacts_service._contacts_next_url == "https://example.com/co/contacts"
    assert (
        contacts_service._contacts_changeset_url == "https://example.com/co/changeset"
    )
    assert contacts_service._contacts_me_card_url == "https://example.com/co/mecard"
    assert contacts_service._contacts is None


@patch("requests.Response")
def test_refresh_client(
    mock_response, contacts_service: ContactsService, mock_session: MagicMock
) -> None:
    """Test the refresh_client method."""
    mock_response.json.return_value = {
        "prefToken": "test_pref_token",
        "syncToken": "test_sync_token",
        "contacts": [{"firstName": "John", "lastName": "Doe"}],
    }
    mock_session.get.return_value = mock_response

    contacts_service.refresh_client()

    mock_session.get.assert_called()
    assert contacts_service._contacts == [
        {
            "firstName": "John",
            "lastName": "Doe",
        }
    ]


@patch("requests.Response")
def test_all_property(
    mock_response, contacts_service: ContactsService, mock_session: MagicMock
) -> None:
    """Test the all property."""
    mock_response.json.return_value = {
        "prefToken": "test_pref_token",
        "syncToken": "test_sync_token",
        "contacts": [{"firstName": "John", "lastName": "Doe"}],
    }
    mock_session.get.return_value = mock_response

    contacts = contacts_service.all

    mock_session.get.assert_called()
    assert contacts == [{"firstName": "John", "lastName": "Doe"}]


@patch("requests.Response")
def test_me_property(
    mock_response, contacts_service: ContactsService, mock_session: MagicMock
) -> None:
    """Test the me property."""
    mock_response.json.return_value = {
        "contacts": [{"firstName": "Jane", "lastName": "Smith", "photo": "photo_url"}]
    }
    mock_session.get.return_value = mock_response

    me_card: MeCard = contacts_service.me

    mock_session.get.assert_called()
    assert isinstance(me_card, MeCard)
    assert me_card.first_name == "Jane"
    assert me_card.last_name == "Smith"
    assert me_card.photo == "photo_url"


def test_me_card_initialization() -> None:
    """Test the initialization of MeCard."""
    data: dict[str, list[dict[str, str]]] = {
        "contacts": [
            {"firstName": "Alice", "lastName": "Johnson", "photo": "photo_url"}
        ]
    }
    me_card = MeCard(data)

    assert me_card.first_name == "Alice"
    assert me_card.last_name == "Johnson"
    assert me_card.photo == "photo_url"
    assert me_card.raw_data == data


def test_me_card_invalid_data() -> None:
    """Test MeCard initialization with invalid data."""
    with pytest.raises(KeyError):
        MeCard({"invalid_key": "value"})



"""Tests for the SSL context configuration in pyicloud.ssl_context."""

import warnings
from typing import Any

import requests
from pytest import MonkeyPatch
from urllib3.exceptions import InsecureRequestWarning

from pyicloud.ssl_context import configurable_ssl_verification


def test_ssl_verification_true(monkeypatch: MonkeyPatch) -> None:
    """Test that SSL verification is enabled by default."""
    called: dict[str, Any] = {}

    def fake_merge(self, url, proxies, stream, verify, cert) -> dict[str, Any]:  # pylint: disable=unused-argument
        called["verify"] = verify
        called["proxies"] = proxies
        return {"verify": verify, "proxies": proxies}

    monkeypatch.setattr(requests.Session, "merge_environment_settings", fake_merge)
    with configurable_ssl_verification():
        session = requests.Session()
        session.merge_environment_settings("https://example.com", {}, False, True, None)
    assert called["verify"] is True
    assert called["proxies"] == {}


def test_ssl_verification_false(monkeypatch: MonkeyPatch) -> None:
    """Test that SSL verification is disabled when verify_ssl=False."""
    called: dict[str, Any] = {}

    def fake_merge(self, url, proxies, stream, verify, cert) -> dict[str, Any]:  # pylint: disable=unused-argument
        called["verify"] = verify
        called["proxies"] = proxies
        return {"verify": verify, "proxies": proxies}

    monkeypatch.setattr(requests.Session, "merge_environment_settings", fake_merge)
    with configurable_ssl_verification(verify_ssl=False):
        session = requests.Session()
        result = session.merge_environment_settings(
            "https://example.com", {}, False, True, None
        )
    assert result["verify"] is False
    assert result["proxies"] == {}


def test_proxy_settings(monkeypatch: MonkeyPatch) -> None:
    """Test that proxy settings are applied correctly."""
    called: dict[str, Any] = {}

    def fake_merge(self, url, proxies, stream, verify, cert) -> dict[str, Any]:  # pylint: disable=unused-argument
        called["verify"] = verify
        called["proxies"] = proxies
        return {"verify": verify, "proxies": proxies}

    monkeypatch.setattr(requests.Session, "merge_environment_settings", fake_merge)
    with configurable_ssl_verification(
        http_proxy="http://proxy", https_proxy="https://proxy"
    ):
        session = requests.Session()
        result = session.merge_environment_settings(
            "https://example.com", {}, False, True, None
        )
    assert result["proxies"] == {"http": "http://proxy", "https": "https://proxy"}


def test_insecure_request_warning(monkeypatch: MonkeyPatch) -> None:
    """Test that InsecureRequestWarning is suppressed when verify_ssl=False."""
    warnings.simplefilter("always")
    monkeypatch.setattr(
        requests.Session, "merge_environment_settings", lambda *a, **kw: {}
    )
    with configurable_ssl_verification(verify_ssl=False):
        with warnings.catch_warnings(record=True) as w:
            warnings.warn("test", InsecureRequestWarning)
            # InsecureRequestWarning should be suppressed
            insecure_warnings: list[warnings.WarningMessage] = [
                warning
                for warning in w
                if issubclass(warning.category, InsecureRequestWarning)
            ]
            assert len(insecure_warnings) == 0, (
                "InsecureRequestWarning should be suppressed"
            )
